from mindsdb_sql_parser.ast import Identifier, Operation, BinaryOperation, BetweenOperation, OrderBy

from mindsdb.api.executor.planner.exceptions import PlanningException


def find_time_filter(op, time_column_name):
    if not op:
        return
    if op.op == 'and':
        left = find_time_filter(op.args[0], time_column_name)
        right = find_time_filter(op.args[1], time_column_name)
        if left and right:
            raise PlanningException('Can provide only one filter by predictor order_by column, found two')

        return left or right
    elif ((isinstance(op.args[0], Identifier) and op.args[0].parts[-1].lower() == time_column_name.lower())
          or (isinstance(op.args[1], Identifier) and op.args[1].parts[-1].lower() == time_column_name.lower())):
        return op


def replace_time_filter(op, time_filter, new_filter):
    if op == time_filter:
        return new_filter
    if isinstance(op, BinaryOperation):
        op.args[0] = replace_time_filter(op.args[0], time_filter, new_filter)
        op.args[1] = replace_time_filter(op.args[1], time_filter, new_filter)
    return op


def find_and_remove_time_filter(op, time_filter):
    if isinstance(op, BinaryOperation) or isinstance(op, BetweenOperation):
        if op == time_filter:
            return None
        elif op.op == 'and':
            # TODO maybe OR operation too?

            # next level
            left_arg = find_and_remove_time_filter(op.args[0], time_filter)
            right_arg = find_and_remove_time_filter(op.args[1], time_filter)

            # if found in one arg return other
            if left_arg is None:
                return right_arg
            if right_arg is None:
                return left_arg

            op.args = [left_arg, right_arg]
            return op

    return op


def validate_ts_where_condition(op, allowed_columns, allow_and=True):
    """Error if the where condition caontains invalid ops, is nested or filters on some column that's not time or partition"""
    if not op:
        return
    allowed_ops = ['and', '>', '>=', '=', '<', '<=', 'between', 'in']
    if not allow_and:
        allowed_ops.remove('and')
    if op.op not in allowed_ops:
        raise PlanningException(
            f'For time series predictors only the following operations are allowed in WHERE: {str(allowed_ops)}, found instead: {str(op)}.')

    for arg in op.args:
        if isinstance(arg, Identifier):
            if arg.parts[-1].lower() not in allowed_columns:
                raise PlanningException(
                    f'For time series predictor only the following columns are allowed in WHERE: {str(allowed_columns)}, found instead: {str(arg)}.')
            # remove alias
            arg.parts = [arg.parts[-1]]

    if isinstance(op.args[0], Operation):
        validate_ts_where_condition(op.args[0], allowed_columns, allow_and=True)
    if isinstance(op.args[1], Operation):
        validate_ts_where_condition(op.args[1], allowed_columns, allow_and=True)


def recursively_check_join_identifiers_for_ambiguity(item, aliased_fields=None):
    if item is None:
        return
    elif isinstance(item, Identifier):
        if len(item.parts) == 1:
            if aliased_fields is not None and item.parts[0] in aliased_fields:
                # is alias
                return
            raise PlanningException(f'Ambigous identifier {str(item)}, provide table name for operations on a join.')
    elif isinstance(item, Operation):
        recursively_check_join_identifiers_for_ambiguity(item.args, aliased_fields=aliased_fields)
    elif isinstance(item, OrderBy):
        recursively_check_join_identifiers_for_ambiguity(item.field, aliased_fields=aliased_fields)
    elif isinstance(item, list):
        for arg in item:
            recursively_check_join_identifiers_for_ambiguity(arg, aliased_fields=aliased_fields)



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log

from mindsdb_sql_parser import ast

logger = log.getLogger(__name__)


def flatten_json(nested_json, parent_key="", separator="."):
    """
    Recursively flattens a nested JSON object into a dictionary with dot notation keys.
    """
    items = []
    for k, v in nested_json.items():
        new_key = f"{parent_key}{separator}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_json(v, new_key, separator=separator).items())
        else:
            items.append((new_key, v))
    return dict(items)


class JiraProjectsTable(APITable):
    """Jira Projects Table implementation"""

    _MAX_API_RESULTS = 100

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Jira "get_all_project_issues" API endpoint
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            jira "get_all_project_issues" matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value

        issues_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] != "key":
                    continue
                if an_order.field.parts[1] in ["reporter", "assignee", "status"]:
                    if issues_kwargs != {}:
                        raise ValueError(
                            "Duplicate order conditions found for reporter,status and assignee"
                        )
                    issues_kwargs["sort"] = an_order.field.parts[1]
                    issues_kwargs["direction"] = an_order.direction
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )
        project = self.handler.connection_data["project"]
        jira_project_df = self.call_jira_api(project)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(jira_project_df) == 0:
            jira_project_df = pd.DataFrame([], columns=selected_columns)
            return jira_project_df

        jira_project_df.columns = self.get_columns()
        for col in set(jira_project_df.columns).difference(set(selected_columns)):
            jira_project_df = jira_project_df.drop(col, axis=1)

        if len(order_by_conditions.get("columns", [])) > 0:
            jira_project_df = jira_project_df.sort_values(
                by=order_by_conditions["columns"],
                ascending=order_by_conditions["ascending"],
            )

        if query.limit:
            jira_project_df = jira_project_df.head(total_results)

        return jira_project_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]
            List of columns
        """
        return [
            "key",
            "summary",
            "status",
            "reporter",
            "assignee",
            "priority",
        ]

    def call_jira_api(self, project):

        jira = self.handler.connect()
        max_records = jira.get_project_issues_count(project)
        max_records = 100
        jql_query = self.handler.construct_jql()
        max_results = self._MAX_API_RESULTS
        start_index = 0
        total = 1
        fields = [
            "key",
            "fields.summary",
            "fields.status.name",
            "fields.reporter.displayName",
            "fields.assignee.displayName",
            "fields.priority.name",
        ]

        all_jira_issues_df = pd.DataFrame(columns=fields)

        while start_index <= total:
            results = self.handler.connect().jql(
                jql_query, start=start_index, limit=max_results
            )
            flattened_data = [flatten_json(item) for item in results["issues"]]
            df = pd.DataFrame(flattened_data)
            df = df[fields]
            start_index += max_results
            total = results["total"]
            all_jira_issues_df = pd.concat([all_jira_issues_df, df], axis=0)

        all_jira_issues_df = all_jira_issues_df.rename(
            columns={
                "key": "key",
                "fields.summary": "summary",
                "fields.reporter.displayName": "reporter",
                "fields.assignee.displayName": "assignee",
                "fields.priority.name": "priority",
                "fields.status.name": "status",
            },
            errors="ignore",
        )

        return all_jira_issues_df



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Impala server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Impala server.',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Impala server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Impala server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Impala server. Must be an integer. Default is 21050'
    }

)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=21050,
    user='USERNAME',
    password='P@55W0Rd',
    database='D4t4bA5e'
)



import lightfm
import numpy as np
import pandas as pd


def get_item_user_idx(args: dict, n_users, n_items, item_ids=None, user_ids=None):
    """
    gets item and user idxs from item_ids and user_ids
    :param args:
    :param n_users:
    :param n_items:
    :param item_ids:
    :param user_ids:
    :return:
    """
    if item_ids and user_ids:
        item_idx = np.array(
            [int(args["item_id_to_idx_map"][item_id]) for item_id in item_ids]
        )
        user_idx = np.array(
            [int(args["user_id_to_idx_map"][user_id]) for user_id in user_ids]
        )

        # repeat each user id index n_items times
        user_idxs = np.repeat(user_idx, n_items)

        # repeat the full list of item indexes n_user times
        item_idxs = np.tile(item_idx, n_users)

    elif item_ids and not user_ids:
        item_idx = np.array(
            [int(args["item_id_to_idx_map"][item_id]) for item_id in item_ids]
        )

        user_idxs = np.repeat([i for i in range(0, n_users)], n_items)
        item_idxs = np.tile(item_idx, n_users)

    elif user_ids and not item_ids:
        user_idx = np.array(
            [int(args["user_id_to_idx_map"][user_id]) for user_id in user_ids]
        )

        user_idxs = np.repeat(user_idx, n_items)
        item_idxs = np.tile([i for i in range(0, n_items)], n_users)

    else:
        user_idxs = np.repeat([i for i in range(0, n_users)], n_items)
        item_idxs = np.tile([i for i in range(0, n_items)], n_users)

    return item_idxs, user_idxs


def get_user_item_recommendations(
    n_users: int, n_items: int, args: dict, item_ids, user_ids, model: lightfm.LightFM
):
    """
    gets N user-item recommendations for a given model
    :param n_users:
    :param n_items:
    :param args:
    :param item_ids:
    :param user_ids:
    :param model:

    :return:
    """
    # get idxs for user-item pairs
    item_idxs, user_idxs = get_item_user_idx(args, n_users, n_items, item_ids, user_ids)

    scores = model.predict(user_ids=user_idxs, item_ids=item_idxs)

    # map scores to user-item pairs, sort by score and return top N recommendations per user
    user_item_recommendations_df = (
        pd.DataFrame({"user_idx": user_idxs, "item_idx": item_idxs, "score": scores})
        .groupby("user_idx")
        .apply(
            lambda x: x.sort_values("score", ascending=False).head(
                args["n_recommendations"]
            )
        )
    )

    # map idxs to item ids and user ids
    user_item_recommendations_df["item_id"] = (
        user_item_recommendations_df["item_idx"]
        .astype("str")
        .map(args["item_idx_to_id_map"])
    )
    user_item_recommendations_df["user_id"] = (
        user_item_recommendations_df["user_idx"]
        .astype("str")
        .map(args["user_idx_to_id_map"])
    )

    return user_item_recommendations_df[["user_id", "item_id", "score"]]


def get_item_item_recommendations(
    model: lightfm.LightFM,
    args: dict,
    item_ids: list = None,
    item_features=None,
) -> pd.DataFrame:
    """
    gets similar items to a given item index inside user-item interaction matrix
    NB by default it won't use item features,however if item features are provided
    it will use them to get similar items

    :param args:
    :param model:
    :param item_ids:
    :param item_features:

    :return:
    """
    # todo make sure its not slow across larger data
    # todo break into smaller functions

    n_recommendations = args["n_recommendations"]

    similar_items_dfs = []

    item_idx_to_id_map = args["item_idx_to_id_map"]

    if item_ids:
        # filter out item_ids that are not in the request
        item_idx_to_id_map = {
            key: val
            for key, val in args["item_idx_to_id_map"].items()
            if val in item_ids
        }

    for item_idx, item_id in item_idx_to_id_map.items():
        # ensure item_idx is int
        item_idx = int(item_idx)

        item_biases, item_representations = model.get_item_representations(
            features=item_features
        )

        # Cosine similarity
        scores = item_representations.dot(item_representations[item_idx, :])

        # normalize
        item_norms = np.sqrt((item_representations * item_representations).sum(axis=1))
        scores /= item_norms

        # ensure n_recommendations is not greater than number of recommendations
        N = min(len(scores), n_recommendations)

        # get the top N items
        best = np.argpartition(scores, -N)

        # sort the scores
        rec = sorted(
            zip(best, scores[best] / item_norms[item_idx]), key=lambda x: -x[1]
        )

        intermediate_df = (
            pd.DataFrame(rec, columns=["item_idx", "score"])
            .tail(-1)  # remove the item itself
            .head(N)
        )

        intermediate_df["item_id_one"] = item_id
        similar_items_dfs.append(intermediate_df)

    similar_items_df = pd.concat(similar_items_dfs, ignore_index=True)
    similar_items_df["item_idx"] = similar_items_df["item_idx"].astype("str")

    similar_items_df["item_id_two"] = similar_items_df["item_idx"].map(
        args["item_idx_to_id_map"]
    )

    return similar_items_df[["item_id_one", "item_id_two", "score"]]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    username={
        "type": ARG_TYPE.STR,
        "description": "DockerHub username",
        "required": True,
        "label": "username",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "DockerHub password",
        "required": True,
        "label": "Api key",
        'secret': True
    }
)

connection_args_example = OrderedDict(
    username="username",
    password="password"
)



__title__ = "MindsDB Confluence handler"
__package_name__ = "mindsdb_confluence_handler"
__version__ = "0.0.1"
__description__ = "MindsDB handler for Confluence"
__author__ = "Balaji Seetharaman"
__github__ = "https://github.com/mindsdb/mindsdb"
__pypi__ = "https://pypi.org/project/mindsdb/"
__license__ = "MIT"
__copyright__ = "Copyright 2023 - mindsdb"



import pandas as pd
import numpy as np

from questdb.ingress import Sender

from mindsdb.integrations.handlers.postgres_handler import Handler as PostgresHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class QuestDBHandler(PostgresHandler):
    """
    This handler handles connection and execution of the QuestDB statements. 
    TODO: check the dialect for questdb
    """
    name = 'questdb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    def get_tables(self):
        """
        List all tabels in QuestDB
        """
        query = "SHOW TABLES"
        response = super().native_query(query)
        return response

    def get_columns(self, table_name):
        """
        List information about the table
        """
        query = f"SELECT * FROM tables() WHERE name='{table_name}';"
        response = super().native_query(query)
        return response

    def qdb_connect(self):
        args = self.connection_args
        conf = f"http::addr={args['host']}:9000;username={args['user']};password={args['password']};"
        return Sender.from_conf(conf)

    def insert(self, table_name: str, df: pd.DataFrame):

        with self.qdb_connect() as sender:
            try:
                # find datetime column
                at_col = None
                for col, dtype in df.dtypes.items():
                    if np.issubdtype(dtype, np.datetime64):
                        at_col = col
                if at_col is None:
                    raise Exception(f'Unable to find datetime column: {df.dtypes}')

                sender.dataframe(df, table_name=table_name, at=at_col)
                response = Response(RESPONSE_TYPE.OK)

            except Exception as e:
                logger.error(f'Error running insert to {table_name} on {self.database}, {e}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )

        return response



from typing import Any, Dict, Union, Optional, List
from uuid import uuid4
import datetime

from langchain_core.callbacks.base import BaseCallbackHandler

from mindsdb.utilities import log
from mindsdb.interfaces.storage import db

logger = log.getLogger(__name__)
logger.setLevel('DEBUG')


class LangfuseCallbackHandler(BaseCallbackHandler):
    """Langchain callback handler that traces tool & chain executions using Langfuse."""

    def __init__(self, langfuse, trace_id: Optional[str] = None, observation_id: Optional[str] = None):
        self.langfuse = langfuse
        self.chain_uuid_to_span = {}
        self.action_uuid_to_span = {}
        # if these are not available, we generate some UUIDs
        self.trace_id = trace_id or uuid4().hex
        self.observation_id = observation_id or uuid4().hex

    def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        """Run when tool starts running."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return
        metadata = {
            'tool_name': serialized.get("name", "tool"),
            'started': datetime.datetime.now().isoformat()
        }
        action_span.update(metadata=metadata)

    def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        """Run when tool ends running."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return
        action_span.update(
            output=output,  # tool output is action output (unless superseded by a global action output)
            metadata={'finished': datetime.datetime.now().isoformat()}
        )

    def on_tool_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when tool errors."""
        parent_run_uuid = kwargs.get('parent_run_id', uuid4()).hex
        action_span = self.action_uuid_to_span.get(parent_run_uuid)
        if action_span is None:
            return
        try:
            error_str = str(error)
        except Exception:
            error_str = "Couldn't get error string."
        action_span.update(metadata={'error_description': error_str})

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        """Run when chain starts running."""
        run_uuid = kwargs.get('run_id', uuid4()).hex

        if serialized is None:
            serialized = {}

        chain_span = self.langfuse.span(
            name=f'{serialized.get("name", "chain")}-{run_uuid}',
            trace_id=self.trace_id,
            parent_observation_id=self.observation_id,
            input=str(inputs)
        )
        self.chain_uuid_to_span[run_uuid] = chain_span

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        chain_uuid = kwargs.get('run_id', uuid4()).hex
        if chain_uuid not in self.chain_uuid_to_span:
            return
        chain_span = self.chain_uuid_to_span.pop(chain_uuid)
        if chain_span is None:
            return
        chain_span.update(output=str(outputs))
        chain_span.end()

    def on_chain_error(self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any) -> Any:
        """Run when chain errors."""
        # Do nothing for now.
        pass

    def on_agent_action(self, action, **kwargs: Any) -> Any:
        """Run on agent action."""
        # Do nothing for now.
        run_uuid = kwargs.get('run_id', uuid4()).hex
        action_span = self.langfuse.span(
            name=f'{getattr(action, "type", "action")}-{getattr(action, "tool", "")}-{run_uuid}',
            trace_id=self.trace_id,
            parent_observation_id=self.observation_id,
            input=str(action)
        )
        self.action_uuid_to_span[run_uuid] = action_span

    def on_agent_finish(self, finish, **kwargs: Any) -> Any:
        """Run on agent end."""
        # Do nothing for now.
        run_uuid = kwargs.get('run_id', uuid4()).hex
        if run_uuid not in self.action_uuid_to_span:
            return
        action_span = self.action_uuid_to_span.pop(run_uuid)
        if action_span is None:
            return
        if finish is not None:
            action_span.update(output=finish)  # supersedes tool output
        action_span.end()

    def auth_check(self):
        if self.langfuse is not None:
            return self.langfuse.auth_check()
        return False


def get_skills(agent: db.Agents) -> List:
    """ Retrieve skills from agent `skills` attribute. Specific to agent endpoints. """
    return [rel.skill.type for rel in agent.skills_relationships]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


creation_args = OrderedDict(
    unify_api_key={
        'type': ARG_TYPE.STR,
        'description': 'Key for Unify API.',
        'required': False,
        'label': 'Unify API key',
        'secret': True
    }
)




class AuthException(Exception):
    def __init__(self, message, auth_url=None):
        super().__init__(message)

        self.auth_url = auth_url


class NoCredentialsException(Exception):
    def __init__(self, message):
        super().__init__(message)



from typing import Union

from mindsdb_sql_parser import parse_sql
from mindsdb.api.executor.planner import utils as planner_utils

from numpy import dtype as np_dtype
from pandas.api import types as pd_types

from mindsdb.api.executor.sql_query import SQLQuery
from mindsdb.api.executor.sql_query.result_set import Column
from mindsdb.api.mysql.mysql_proxy.utilities.lightwood_dtype import dtype
from mindsdb.api.executor.command_executor import ExecuteCommands
from mindsdb.api.mysql.mysql_proxy.utilities import SqlApiException
from mindsdb.api.postgres.postgres_proxy.postgres_packets.postgres_fields import POSTGRES_TYPES
from mindsdb.utilities import log


class Executor:
    def __init__(self, session, proxy_server, charset=None):
        self.session = session
        self.server = proxy_server

        self.logger = log.getLogger(__name__)
        self.charset = charset or "utf8"
        self.query = None
        self.columns = []
        self.params = []
        self.data = None
        self.server_status = None
        self.state_track = None
        self.is_executed = False

        self.sql = ""
        self.sql_lower = ""

        self.command_executor = ExecuteCommands(self.session)

    def parse(self, sql: Union[str, bytes]):
        self.logger.info("%s.parse: sql - %s", self.__class__.__name__, sql)
        if type(sql) == bytes:
            sql = sql.decode(encoding=self.charset)
        self.sql = sql
        sql_lower = self.sql.lower()
        self.sql_lower = sql_lower.replace("`", "")

        try:
            self.query = parse_sql(sql)
        except Exception as mdb_error:
            # not all statements are parsed by parse_sql
            self.logger.warning('Failed to parse SQL query')
            self.logger.debug(f'Query that cannot be parsed: {sql}')

            raise SqlApiException(
                f"The SQL statement cannot be parsed - {sql}: {mdb_error}"
            ) from mdb_error

    def stmt_execute(self, param_values):
        if self.is_executed:
            return

        # fill params
        self.query = planner_utils.fill_query_params(self.query, param_values)

        # execute query
        self.do_execute()

    def execute_external(self, sql):
        return None

    def query_execute(self, sql):
        self.logger.info("%s.query_execute: sql - %s", self.__class__.__name__, sql)
        resp = self.execute_external(sql)
        if resp is not None:
            # is already executed
            self.is_executed = True
            return

        self.parse(sql)
        self.do_execute()

    def do_execute(self):
        self.logger.info("%s.do_execute", self.__class__.__name__)
        # it can be already run at prepare state
        if self.is_executed:
            return

        ret = self.command_executor.execute_command(self.query)

        self.is_executed = True

        if ret.data is not None:
            self.data = ret.data.to_lists()
            self.columns = ret.data.columns

        self.state_track = ret.state_track

    def _to_json(self):
        params = {
            "columns": self.to_postgres_columns(self.columns),
            "params": self.to_postgres_columns(self.params),
            "data": self.data,
            "state_track": self.state_track,
            "server_status": self.server_status,
            "is_executed": self.is_executed,
            "session": self.session.to_json()

        }
        return params

    def to_postgres_columns(self, columns):

        result = []

        database = (
            None if self.session.database == "" else self.session.database.lower()
        )
        for column_record in columns:

            field_type = column_record.type

            column_type = POSTGRES_TYPES.VARCHAR
            # is already in mysql protocol type?
            if isinstance(field_type, int):
                column_type = POSTGRES_TYPES.INT
            # pandas checks
            elif isinstance(field_type, np_dtype):
                if pd_types.is_integer_dtype(field_type):
                    column_type = POSTGRES_TYPES.LONG
                elif pd_types.is_numeric_dtype(field_type):
                    column_type = POSTGRES_TYPES.DOUBLE
                elif pd_types.is_datetime64_any_dtype(field_type):
                    column_type = POSTGRES_TYPES.DATETIME
            # lightwood checks
            elif field_type == dtype.date:
                column_type = POSTGRES_TYPES.DATE
            elif field_type == dtype.datetime:
                column_type = POSTGRES_TYPES.DATETIME
            elif field_type == dtype.float:
                column_type = POSTGRES_TYPES.DOUBLE
            elif field_type == dtype.integer:
                column_type = POSTGRES_TYPES.LONG

            if "()" in column_record.alias:
                column_record.alias = column_record.alias.strip("()")
            if "()" in column_record.name:
                column_record.name = column_record.name.strip("()")

            result.append(
                {
                    "database": column_record.database or database,
                    #  TODO add 'original_table'
                    "table_name": column_record.table_name,
                    "name": column_record.name,
                    "alias": column_record.alias or column_record.name,
                    # NOTE all work with text-type, but if/when wanted change types to real,
                    # it will need to check all types casts in BinaryResultsetRowPacket
                    "type": column_type,
                }
            )
        return result

    def change_default_db(self, new_db):
        self.command_executor.change_default_db(new_db)

    def stmt_prepare(self, sql):
        # Returns True if ready for query afterwards.
        # Check if execute external here
        self.parse(sql)
        params = planner_utils.get_query_params(self.query)
        if len(params) == 0:
            pass
        #    self.do_execute()
        #    return True
        else:
            # plan query

            sqlquery = SQLQuery(self.query, session=self.session, execute=False)

            sqlquery.prepare_query()

            self.params = [
                Column(
                    alias=p.value,
                    type="str",
                    name=p.value,
                )
                for p in params
            ]
            self.columns = sqlquery.columns_list



import os
from typing import Iterable, Callable
import pandas as pd

from mindsdb.utilities.config import Config
from mindsdb.utilities.context_executor import execute_in_threads


def process_dataframe_in_partitions(df: pd.DataFrame, callback: Callable, partition_size: int) -> Iterable:
    """
    Splits dataframe into partitions and apply callback on each partition

    :param df: input dataframe
    :param callback: function to apply on each partition
    :param partition_size: size of each partition
    :return: yield results
    """

    # tasks
    def split_data_f(df):
        chunk = 0
        while chunk * partition_size < len(df):
            # create results with partition
            df1 = df.iloc[chunk * partition_size: (chunk + 1) * partition_size]
            chunk += 1
            yield [df1]

    tasks = split_data_f(df)

    # workers count
    is_cloud = Config().is_cloud
    if is_cloud:
        max_threads = int(os.getenv('MINDSDB_MAX_PARTITIONING_THREADS', 10))
    else:
        max_threads = os.cpu_count() - 2

    # don't exceed chunk_count
    chunk_count = int(len(df) / partition_size)
    max_threads = min(max_threads, chunk_count)

    if max_threads < 1:
        max_threads = 1

    if max_threads == 1:
        # don't spawn threads

        for task in tasks:
            yield callback(*task)

    else:
        for result in execute_in_threads(callback, tasks, thread_count=max_threads):
            yield result



from typing import List
import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql_parser.ast.select.constant import Constant
import json
import re

langCodes = ["ar", "cs", "da", "de", "en", "es", "et", "fi", "fr", "hu", "id", "it", "ja", "ko",
             "nl", "no", "pl", "pt", "pt-BR", "ro", "rs", "ru", "sl", "sq", "sv", "tr", "uk",
             "vi", "zh", "zh-TW"]


class StatusPages(APITable):

    # table name in the database
    name = 'status_pages'

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id from query
        _id = None
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError

        # Get column names from query
        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        # 'id' needs to selected when searching with 'id'
        temp_selected_columns = selected_columns
        if _id and 'id' not in selected_columns:
            selected_columns = ['id'] + selected_columns

        # Get limit from query
        limit = query.limit.value if query.limit else 20
        total_results = limit

        page_no = 1  # default page no
        result_df = pd.DataFrame(columns=selected_columns)

        # call instatus api and get the response as pd.DataFrame
        while True:
            df = self.handler.call_instatus_api(endpoint='/v2/pages', params={'page': page_no, 'per_page': 100})
            if len(df) == 0 or limit <= 0:
                break
            else:
                result_df = pd.concat([result_df, df[selected_columns]], ignore_index=True)

            page_no += 1
            limit -= len(df)

        # select columns from pandas data frame df
        if result_df.empty:
            result_df = pd.DataFrame(columns=selected_columns)
        elif _id:
            result_df = result_df[result_df['id'] == _id]

        # delete 'id' column if 'id' not present in temp_selected_columns
        if 'id' not in temp_selected_columns and 'id' in selected_columns:
            result_df = result_df.drop('id', axis=1)

        return result_df.head(n=total_results)

    def insert(self, query: ast.Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        data = {}
        for column, value in zip(query.columns, query.values[0]):
            if isinstance(value, str):
                try:
                    value = json.loads(value)
                except json.JSONDecodeError:
                    if value == 'True':
                        value = True
                    elif value == 'False':
                        value = False
            data[column.name] = value
        self.handler.call_instatus_api(endpoint='/v1/pages', method='POST', json_data=data)

    def update(self, query: ast.Update) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id from query
        _id = None
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError

        data = {}
        for key, value in query.update_columns.items():
            if isinstance(value, Constant):
                if key == 'components':
                    data[key] = json.loads(value.value)
                else:
                    data[key] = value.value

        if 'components' in data and isinstance(data['components'], str):
            data['components'] = json.loads(data['components'])

        self.handler.call_instatus_api(endpoint=f'/v2/{_id}', method='PUT', json_data=data)

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """
        return [
            "id",
            "subdomain",
            "name",
            "workspaceId",
            "logoUrl",
            "faviconUrl",
            "websiteUrl",
            "customDomain",
            "publicEmail",
            "twitter",
            "status",
            "subscribeBySms",
            "sendSmsNotifications",
            "language",
            "useLargeHeader",
            "brandColor",
            "okColor",
            "disruptedColor",
            "degradedColor",
            "downColor",
            "noticeColor",
            "unknownColor",
            "googleAnalytics",
            "smsService",
            "htmlInMeta",
            "htmlAboveHeader",
            "htmlBelowHeader",
            "htmlAboveFooter",
            "htmlBelowFooter",
            "htmlBelowSummary",
            "uptimeDaysDisplay",
            "uptimeOutageDisplay",
            "launchDate",
            "cssGlobal",
            "onboarded",
            "createdAt",
            "updatedAt"
        ]


class Components(APITable):

    # table name in the database
    name = 'components'

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): SQL query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """
        conditions = extract_comparison_conditions(query.where)

        if len(conditions) == 0:
            raise Exception('WHERE clause is required')

        # Get page id and component id from query
        pageId = None
        componentId = None
        for condition in conditions:
            if condition[1] == 'page_id' and condition[0] == '=':
                pageId = condition[2]

            if condition[1] == 'component_id' and condition[0] == '=':
                componentId = condition[2]

        # Get column names from query
        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        limit = query.limit.value if query.limit else None
        if componentId:
            # Call instatus API and get the response as pd.DataFrame
            df = self.handler.call_instatus_api(endpoint=f'/v1/{pageId}/components/{componentId}')
            for langCode in langCodes:
                try:
                    df[f"translations_name_in_{langCode}"] = df["translations"].apply(lambda x: x.get("name", None)).apply(lambda x: x.get(langCode, None))
                    df[f"translations_desc_in_{langCode}"] = df["translations"].apply(lambda x: x.get("description", None)).apply(lambda x: x.get(langCode, None))
                except AttributeError:
                    df[f"translations_name_in_{langCode}"] = None
                    df[f"translations_desc_in_{langCode}"] = None
            df = df.drop(columns=["translations"])

            result_df = df[selected_columns]
        else:
            # Call instatus API and get the response as pd.DataFrame
            page_size = 100
            # Calculate the number of pages required
            page_count = (limit + page_size - 1) // page_size if limit else 1
            result_df = pd.DataFrame(columns=selected_columns)

            # Call instatus API and get the response as pd.DataFrame for each page
            for page in range(1, page_count + 1):
                current_page_size = min(page_size, limit) if limit else page_size

                df = self.handler.call_instatus_api(endpoint=f'/v1/{pageId}/components', params={'page': page, 'per_page': current_page_size})
                # Break if no more data is available or limit is reached
                if len(df) == 0 or (limit and limit <= 0) or limit == 0:
                    break
                ''' Add translations_name_in_{langCode} and translations_desc_in_{langCode} columns to the dataframe'''
                for i in range(len(df)):
                    for langCode in langCodes:
                        try:
                            df.at[i, f"translations_name_in_{langCode}"] = df.at[i, "translations"].get("name", {}).get(langCode, None)
                            df.at[i, f"translations_desc_in_{langCode}"] = df.at[i, "translations"].get("description", {}).get(langCode, None)
                        except AttributeError:
                            df.at[i, f"translations_name_in_{langCode}"] = None
                            df.at[i, f"translations_desc_in_{langCode}"] = None

                # Drop the 'translations' column
                df = df.drop(columns=["translations"])
                # Concatenate the dataframes
                result_df = pd.concat([result_df, df[selected_columns]], ignore_index=True)

                if limit:
                    limit -= len(df)

        return result_df

    def insert(self, query: ast.Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        data = {'translations': {
            "name": {},
            "description": {}
        }}

        for column, value in zip(query.columns, query.values[0]):
            if isinstance(value, Constant):
                data[column.name] = json.loads(value.value) if column.name == 'translations' else value.value
            elif isinstance(value, str):
                try:
                    if re.match(r'^translations_name_in_[a-zA-Z\-]+$', column.name):
                        lang_code = column.name.split('_')[-1]
                        if lang_code not in langCodes:
                            raise Exception(f'Invalid language code {lang_code}')
                        data['translations']['name'][lang_code] = value
                    elif re.match(r'^translations_desc_in_[a-zA-Z\-]+$', column.name):
                        lang_code = column.name.split('_')[-1]
                        if lang_code not in langCodes:
                            raise Exception(f'Invalid language code {lang_code}')
                        data['translations']['description'][lang_code] = value
                    else:
                        data[column.name] = json.loads(value)
                except json.JSONDecodeError:
                    data[column.name] = True if value == 'True' else (False if value == 'False' else value)

        page_id = data.pop('page_id', None)

        if page_id is not None:
            self.handler.call_instatus_api(endpoint=f'/v1/{page_id}/components', method='POST', json_data=data)

    def update(self, query: ast.Update) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id and component id from query
        pageId = None
        componentId = None
        for condition in conditions:
            if condition[1] == 'page_id' and condition[0] == '=':
                pageId = condition[2]
            elif condition[1] == 'component_id' and condition[0] == '=':
                componentId = condition[2]
            else:
                raise Exception("page_id and component_id both are required")

        data = {'translations': {
            "name": {},
            "description": {}
        }}
        for key, value in query.update_columns.items():
            if isinstance(value, Constant):
                if re.match(r'^translations_name_in_[a-zA-Z\-]+$', key):
                    lang_code = key.split('_')[-1]
                    if lang_code not in langCodes:
                        raise Exception(f'Invalid language code {lang_code}')
                    data['translations']['name'][lang_code] = value.value
                elif re.match(r'^translations_desc_in_[a-zA-Z\-]+$', key):
                    lang_code = key.split('_')[-1]
                    if lang_code not in langCodes:
                        raise Exception(f'Invalid language code {lang_code}')
                    data['translations']['description'][lang_code] = value.value
                else:
                    data[key] = value.value
        self.handler.call_instatus_api(endpoint=f'/v1/{pageId}/components/{componentId}', method='PUT', json_data=data)

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """
        return [
            "id",
            "name",
            "nameTranslationId",
            "description",
            "descriptionTranslationId",
            "status",
            "order",
            "showUptime",
            "createdAt",
            "updatedAt",
            "archivedAt",
            "siteId",
            "uniqueEmail",
            "oldGroup",
            "groupId",
            "isParent",
            "isCollapsed",
            "monitorId",
            "nameHtml",
            "nameHtmlTranslationId",
            "descriptionHtml",
            "descriptionHtmlTranslationId",
            "isThirdParty",
            "thirdPartyStatus",
            "thirdPartyComponentId",
            "thirdPartyComponentServiceId",
            "importedFromStatuspage",
            "startDate",
            "group",
        ] + [f'translations_name_in_{langCode}' for langCode in langCodes] + [f'translations_desc_in_{langCode}' for langCode in langCodes]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Kinetica server.',
        'required': True,
        'label': 'User',
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Kinetica server.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Kinetica server.',
        'required': True,
        'label': 'Database',
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Kinetica server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.',
        'required': True,
        'label': 'Host',
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Kinetica server. Must be an integer.',
        'required': True,
        'label': 'Port',
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'The schema in which objects are searched first.',
        'required': False,
        'label': 'Schema',
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': 'sslmode that will be used for connection.',
        'required': False,
        'label': 'sslmode',
    },
)

connection_args_example = OrderedDict(
    host='127.0.0.1', port=5432, user='root', password='password', database='database'
)



"""predictor_integration_null

Revision ID: ee63d868fa84
Revises: b6d0a47294ac
Create Date: 2023-02-17 13:48:49.464644

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa


# revision identifiers, used by Alembic.
revision = 'ee63d868fa84'
down_revision = 'b6d0a47294ac'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=True
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.alter_column(
            'integration_id',
            existing_type=sa.INTEGER(),
            nullable=False
        )

    # ### end Alembic commands ###



import unittest

from mindsdb.integrations.handlers.mendeley_handler.mendeley_handler import MendeleyHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE

class MendeleyHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data":{"client_id":15253,"client_secret":"BxmSvbrRW5iYEIQR"}
        }
        cls.handler = MendeleyHandler("test_mendeley_handler",**cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_select(self):
        query = "SELECT * FROM catalog_search_data WHERE doi='10.1111/joim.13091'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_select(self):
        query = "SELECT * FROM catalog_search_data WHERE id='af1a0408-7409-3a8b-ad91-8accd4f8849a'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_4_select(self):
        query = "SELECT * FROM catalog_search_data WHERE title='The American Mineralogist crystal structure database'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_5_select(self):
        query = "SELECT * FROM catalog_search_data WHERE id='8e86b541-84fd-30ef-9eed-e8c9af847ca3' AND doi='10.1093/ajcn/77.1.71'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_6_select(self):
        query = "SELECT * FROM catalog_search_data WHERE issn='15570878'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_7_select(self):
        query = "SELECT * FROM catalog_search_data WHERE source='American Journal of Clinical Nutrition'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_8_select(self):
        query = "SELECT * FROM catalog_search_data WHERE source='American Journal of Clinical Nutrition'AND id='4eeda257-8db4-3dad-80c8-6912356d3887'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_9_select(self):
        query = "SELECT * FROM catalog_search_data WHERE source='American Journal of Clinical Nutrition'AND max_year='2020'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE
    
    def test_10_call_mendeley_api_invalid_method_name(self):
        with self.assertRaises(NotImplementedError):
            res = self.handler.call_mendeley_api("method1",None)

    def test_11_select_invalid_condition_name(self):
        with self.assertRaises(NotImplementedError):
            query = "SELECT * FROM catalog_search_data WHERE name='American Journal of Clinical Nutrition'"
            result = self.handler.native_query(query)

    def test_12_select_invalid_operator(self):
        with self.assertRaises(NotImplementedError):
            query = "SELECT * FROM catalog_search_data WHERE source>'American Journal of Clinical Nutrition'AND max_year='2020' "
            result = self.handler.native_query(query)
    
    def test_13_select_invalid_column_name(self):
        with self.assertRaises(KeyError):
            query = "SELECT name FROM catalog_search_data WHERE source='American Journal of Clinical Nutrition'AND max_year='2020' "
            result = self.handler.native_query(query)
    
    def test_14_get_columns(self):
        columns = self.handler.catalog_search_data.get_columns()

        expected_columns = [

            'title',
            'type',
            'source',
            'year',
            'pmid',
            'sgr',
            'issn',
            'scopus',
            'doi',
            'pui',
            'authors',
            'keywords',
            'link',
            'id'
        ]

        self.assertListEqual(columns, expected_columns)
    
    def test_15_select_invalid_condition_name(self):
        with self.assertRaises(ValueError):
            query = "SELECT * FROM catalog_search_data"
            result = self.handler.native_query(query)

if __name__ == "__main__":
    unittest.main()



from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.handlers.pypi_handler.api import PyPI
from mindsdb.integrations.handlers.pypi_handler.pypi_tables import (
    PyPIOverallTable,
    PyPIPythonMajorTable,
    PyPIPythonMinorTable,
    PyPIRecentTable,
    PyPISystemTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse


class PyPIHandler(APIHandler):
    def __init__(self, name: str, **kwargs) -> None:
        """initializer method

        Args:
            name (str): handler's name
        """
        super().__init__(name)

        self.connection = None
        self.is_connected = False

        _tables = [
            PyPIOverallTable,
            PyPIPythonMajorTable,
            PyPIPythonMinorTable,
            PyPIRecentTable,
            PyPISystemTable,
        ]

        for Table in _tables:
            self._register_table(Table.name, Table(self))

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)
        checking = PyPI.is_connected()
        if checking["status"]:
            response.success = True
        else:
            response.error_message = checking["message"]

        self.is_connected = True

        return response

    def connect(self) -> PyPI:
        """making the connectino object

        Returns:
            PyPI: pypi class as the returned value
        """
        self.connection = PyPI
        return self.connection

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



import copy

import numpy as np

from mindsdb_sql_parser.ast import (
    Identifier, BinaryOperation, Constant
)
from mindsdb.api.executor.planner.steps import (
    JoinStep,
)
from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.utilities.sql import query_df_with_type_infer_fallback
from mindsdb.api.executor.exceptions import NotSupportedYet

from .base import BaseStepCall


class JoinStepCall(BaseStepCall):

    bind = JoinStep

    def call(self, step):
        left_data = self.steps_data[step.left.step_num]
        right_data = self.steps_data[step.right.step_num]

        if right_data.is_prediction or left_data.is_prediction:
            # ignore join condition, use row_id
            l_row_ids = left_data.find_columns('__mindsdb_row_id')
            r_row_ids = right_data.find_columns('__mindsdb_row_id')

            if len(l_row_ids) == 0:
                if len(r_row_ids) == 0:
                    raise RuntimeError('Unable to find row id')
                else:
                    # copy from right to left
                    idx = right_data.get_col_index(r_row_ids[0])
                    left_data.set_column_values('__mindsdb_row_id', right_data.get_column_values(idx))
                    l_row_ids = left_data.find_columns('__mindsdb_row_id')
            elif len(r_row_ids) == 0:
                # copy from left to right
                idx = left_data.get_col_index(l_row_ids[0])
                right_data.set_column_values('__mindsdb_row_id', left_data.get_column_values(idx))
                r_row_ids = right_data.find_columns('__mindsdb_row_id')

            a_row_id = l_row_ids[0].get_hash_name(prefix='A')
            b_row_id = r_row_ids[0].get_hash_name(prefix='B')

            join_condition = f'table_a.{a_row_id} = table_b.{b_row_id}'

            join_type = step.query.join_type.lower()
            if join_type == 'join':
                # join type is not specified. using join to prediction data
                if left_data.is_prediction:
                    join_type = 'left join'
                elif right_data.is_prediction:
                    join_type = 'right join'
        else:
            def adapt_condition(node, **kwargs):
                if not isinstance(node, Identifier) or len(node.parts) != 2:
                    return

                table_alias, alias = node.parts
                cols = left_data.find_columns(alias, table_alias)
                if len(cols) == 1:
                    col_name = cols[0].get_hash_name(prefix='A')
                    return Identifier(parts=['table_a', col_name])

                cols = right_data.find_columns(alias, table_alias)
                if len(cols) == 1:
                    col_name = cols[0].get_hash_name(prefix='B')
                    return Identifier(parts=['table_b', col_name])

            if step.query.condition is None:
                # prevent memory overflow
                if len(left_data) * len(right_data) < 10 ** 7:
                    step.query.condition = BinaryOperation(op='=', args=[Constant(0), Constant(0)])
                else:
                    raise NotSupportedYet('Unable to join table without condition')

            condition = copy.deepcopy(step.query.condition)
            query_traversal(condition, adapt_condition)

            join_condition = SqlalchemyRender('postgres').get_string(condition)
            join_type = step.query.join_type

        table_a, names_a = left_data.to_df_cols(prefix='A')
        table_b, names_b = right_data.to_df_cols(prefix='B')

        query = f"""
            SELECT * FROM table_a {join_type} table_b
            ON {join_condition}
        """
        resp_df, _description = query_df_with_type_infer_fallback(query, {
            'table_a': table_a,
            'table_b': table_b
        })

        resp_df.replace({np.nan: None}, inplace=True)

        names_a.update(names_b)
        data = ResultSet().from_df_cols(resp_df, col_names=names_a)

        for col in data.find_columns('__mindsdb_row_id'):
            data.del_column(col)

        return data



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import struct

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    DEFAULT_CAPABILITIES,
    NULL_VALUE,
    ONE_BYTE_ENC,
    THREE_BYTE_ENC,
    TWO_BYTE_ENC,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

NULL_VALUE_INT = ord(NULL_VALUE)


class Datum:
    __slots__ = ['value', 'var_type', 'var_len']

    def __init__(self, var_type, value=None, var_len=None):
        # TODO other types: float, timestamp
        self.value = b""

        if var_len is None:
            idx = var_type.find('<')
            var_len = var_type[idx + 1: -1]
            var_type = var_type[: idx]
        self.var_type = var_type
        self.var_len = var_len

        if value is not None:
            self.set(value)

    def set(self, value):
        self.value = value

    def setFromBuff(self, buff):
        if self.var_len == "lenenc":
            start = 1
            ln_enc = buff[0]
            if int(ln_enc) <= ONE_BYTE_ENC[0]:
                start = 0
                end = 1
            elif int(ln_enc) == TWO_BYTE_ENC[0]:
                end = 3
            elif int(ln_enc) == THREE_BYTE_ENC[0]:
                end = 4
            elif ln_enc:
                end = 9

            num_str = buff[start:end]
            if end > 9:
                logger.error("Cant decode integer greater than 8 bytes")
                return buff[end - 1 :]  # noqa: E203

            for j in range(8 - (end - start)):
                num_str += b"\0"

            if self.var_type == "int":
                self.value = struct.unpack("i", num_str)
                return buff[end:]

            if self.var_type in ["byte", "string"]:
                length = struct.unpack("Q", num_str)[0]
                self.value = buff[end : (length + end)]  # noqa: E203
                return buff[(length + end) :]  # noqa: E203

        if self.var_len == "EOF":
            length = len(buff)
            self.var_len = str(length)
            self.value = buff
            return ""
        else:
            length = self.var_len

        if self.var_type == "string" and self.var_len == "NUL":
            for j, x in enumerate(buff):
                if int(x) == 0:
                    length = j + 1
                    break

        length = int(length)
        if self.var_type in ["byte", "string"]:
            end = length
            self.value = buff[:end]
        else:  # if its an integer
            end = length
            num_str = buff[:end]
            if end > 8:
                logger.error("cant decode integer greater than 8 bytes")
                return buff[end:]
            for j in range(8 - end):
                num_str += b"\0"
            self.value = struct.unpack("Q", num_str)[0]
        if str(self.var_len) == "NUL":
            self.value = self.value[:-1]
        return buff[end:]

    @classmethod
    def serialize_int(cls, value):
        if value is None:
            return NULL_VALUE

        byte_count = -(value.bit_length() // (-8))

        if byte_count == 0:
            return b"\0"
        if value < NULL_VALUE_INT:
            return struct.pack("B", value)
        if value >= NULL_VALUE_INT and byte_count <= 2:
            return TWO_BYTE_ENC + struct.pack("H", value)
        if byte_count <= 3:
            return THREE_BYTE_ENC + struct.pack("i", value)[:3]
        if byte_count <= 8:
            return THREE_BYTE_ENC + struct.pack("Q", value)

    def toStringPacket(self):
        return self.get_serializer()(self.value)

    def get_serializer(self):
        if self.var_type == "string":
            if self.var_len == "lenenc":
                if isinstance(self.value, bytes):
                    return self.serialize_bytes
                return self.serialize_str
            if self.var_len == "EOF":
                return self.serialize_str_eof
            if self.var_len == "NUL":
                return lambda v: bytes(v, "utf-8") + struct.pack("b", 0)
            if self.var_len == "packet":
                return lambda v: v.get_packet_string()
            else:
                return lambda v: struct.pack(self.var_len + "s", bytes(v, "utf-8"))[
                    :int(self.var_len)
                ]

        if self.var_type == "int":
            if self.var_len == "lenenc":
                return self.serialize_int
            else:
                return lambda v: struct.pack("Q", v)[:int(self.var_len)]

    @classmethod
    def serialize_str_eof(cls, value):
        length = len(value)
        var_len = length
        if length == 0:
            return b""
        else:
            return struct.pack(
                "{len}s".format(len=var_len), bytes(value, "utf-8")
            )[:length]

    # def serialize_obj(self, value):
    #     return self.serialize_str(str(value))

    @classmethod
    def serialize_str(cls, value):
        return cls.serialize_bytes(value.encode("utf8"))

    @classmethod
    def serialize_bytes(cls, value):

        val_len = len(value)

        if val_len == 0:
            return b"\0"

        if val_len < NULL_VALUE_INT:
            return struct.pack("B", val_len) + value

        byte_count = -(val_len.bit_length() // (-8))
        if byte_count <= 2:
            return (
                TWO_BYTE_ENC
                + struct.pack("H", val_len)
                + value
            )
        if byte_count <= 3:
            return (
                THREE_BYTE_ENC
                + struct.pack("i", val_len)[:3]
                + value
            )
        if byte_count <= 8:
            return (
                THREE_BYTE_ENC
                + struct.pack("Q", val_len)
                + value
            )


def test():
    import pprint

    u = Datum("int<8>", DEFAULT_CAPABILITIES >> 16)
    pprint.pprint(u.toStringPacket())


# only run the test if this file is called from debugger
if __name__ == "__main__":
    test()



import os
import sys
import json
import argparse
import datetime
from pathlib import Path
from copy import deepcopy
import multiprocessing as mp

from appdirs import user_data_dir

# NOTE do not `import from mindsdb` here


def _merge_key_recursive(target_dict, source_dict, key):
    if key not in target_dict:
        target_dict[key] = source_dict[key]
    elif not isinstance(target_dict[key], dict) or not isinstance(source_dict[key], dict):
        target_dict[key] = source_dict[key]
    else:
        for k in list(source_dict[key].keys()):
            _merge_key_recursive(target_dict[key], source_dict[key], k)


def _merge_configs(original_config: dict, override_config: dict) -> dict:
    for key in list(override_config.keys()):
        _merge_key_recursive(original_config, override_config, key)
    return original_config


def create_data_dir(path: Path) -> None:
    """Create a directory and checks that it is writable.

    Args:
        path (Path): path to create and check

    Raises:
        NotADirectoryError: if path exists, but it is not a directory
        PermissionError: if path exists/created, but it is not writable
        Exception: if directory could not be created
    """
    if path.exists() and not path.is_dir():
        raise NotADirectoryError(f"The path is not a directory: {path}")

    try:
        path.mkdir(mode=0o777, exist_ok=True, parents=True)
    except Exception as e:
        raise Exception(
            "MindsDB storage directory could not be created"
        ) from e

    if not os.access(path, os.W_OK):
        raise PermissionError(
            f"The directory is not allowed for writing: {path}"
        )


class Config:
    """Application config. Singletone, initialized just once. Re-initialyze if `config.auto.json` is changed.
    The class loads multiple configs and merge then in one. If a config option defined in multiple places (config file,
    env var, cmd arg, etc), then it will be resolved in following order of priority:
     - default config values (lowest priority)
     - `config.json` provided by the user
     - `config.auto.json`
     - config values collected from env vars
     - values from cmd args (most priority)

    Attributes:
        __instance (Config): instance of 'Config' to make it singleton
        _config (dict): application config, the result of merging other configs
        _user_config (dict): config provided by the user (usually with cmd arg `--config=config.json`)
        _env_config (dict): config collected from different env vars
        _auto_config (dict): config that is editd by the app itself (e.g. when you change values in GUI)
        _default_config (dict): config with default values
        config_path (Path): path to the `config.json` provided by the user
        storage_root_path (Path): path to storage root folder
        auto_config_path (Path): path to `config.auto.json`
        auto_config_mtime (float): mtime of `config.auto.json` when it was loaded to `self._auto_config`
        _cmd_args (argparse.Namespace): cmd args
        use_docker_env (bool): is the app run in docker env
    """
    __instance: 'Config' = None

    _config: dict = None
    _user_config: dict = None
    _env_config: dict = None
    _auto_config: dict = None
    _default_config: dict = None
    config_path: Path = None
    storage_root_path: Path = None
    auto_config_path: Path = None
    auto_config_mtime: float = 0
    _cmd_args: argparse.Namespace = None
    use_docker_env: bool = os.environ.get('MINDSDB_DOCKER_ENV', False) is not False

    def __new__(cls, *args, **kwargs) -> 'Config':
        """Make class singletone and initialize config.
        """
        if cls.__instance is not None:
            return cls.__instance

        self = super().__new__(cls, *args, **kwargs)
        cls.__instance = self

        self.fetch_user_config()

        # region determine root path
        if self.storage_root_path is None:
            if isinstance(os.environ.get('MINDSDB_STORAGE_DIR'), str):
                self.storage_root_path = os.environ['MINDSDB_STORAGE_DIR']
            elif 'root' in self._user_config.get('paths', {}):
                self.storage_root_path = self.user_config['paths']['root']
            else:
                self.storage_root_path = os.path.join(
                    user_data_dir("mindsdb", "mindsdb"),
                    "var/"
                )
            self.storage_root_path = Path(self.storage_root_path)
            create_data_dir(self.storage_root_path)
        # endregion

        # region prepare default config
        api_host = "127.0.0.1" if not self.use_docker_env else "0.0.0.0"
        self._default_config = {
            'permanent_storage': {
                'location': 'absent'
            },
            "storage_db": (
                'sqlite:///'
                + str(self.storage_root_path / 'mindsdb.sqlite3.db')
                + '?check_same_thread=False&timeout=30'
            ),
            'paths': {
                'root': self.storage_root_path,
                'content': self.storage_root_path / 'content',
                'storage': self.storage_root_path / 'storage',
                'static': self.storage_root_path / 'static',
                'tmp': self.storage_root_path / 'tmp',
                'log': self.storage_root_path / 'log',
                'cache': self.storage_root_path / 'cache',
                'locks': self.storage_root_path / 'locks',
            },
            'auth': {
                'http_auth_enabled': False,
                "http_permanent_session_lifetime": datetime.timedelta(days=31)
            },
            "logging": {
                "handlers": {
                    "console": {
                        "enabled": True,
                        "level": "INFO"     # MINDSDB_CONSOLE_LOG_LEVEL or MINDSDB_LOG_LEVEL (obsolete)
                    },
                    "file": {
                        "enabled": False,
                        "level": "INFO",    # MINDSDB_FILE_LOG_LEVEL
                        "filename": "app.log",
                        "maxBytes": 1 << 19,    # 0.5 Mb
                        "backupCount": 3
                    }
                }
            },
            "gui": {
                "autoupdate": True
            },
            "debug": False,
            "environment": "local",
            "integrations": {},
            "api": {
                "http": {
                    "host": api_host,
                    "port": "47334",
                    "restart_on_failure": True,
                    "max_restart_count": 1,
                    "max_restart_interval_seconds": 60,
                    "server": {
                        "type": "waitress",     # MINDSDB_HTTP_SERVER_TYPE MINDSDB_DEFAULT_SERVER
                        "config": {
                            "threads": 16,
                            "max_request_body_size": (1 << 30) * 10,    # 10GB
                            "inbuf_overflow": (1 << 30) * 10
                        }
                    }
                },
                "mysql": {
                    "host": api_host,
                    "password": "",
                    "port": "47335",
                    "database": "mindsdb",
                    "ssl": True,
                    "restart_on_failure": True,
                    "max_restart_count": 1,
                    "max_restart_interval_seconds": 60
                },
                "mongodb": {
                    "host": api_host,
                    "port": "47336",
                    "database": "mindsdb"
                },
                "postgres": {
                    "host": api_host,
                    "port": "55432",
                    "database": "mindsdb"
                }
            },
            "cache": {
                "type": "local"
            },
            'ml_task_queue': {
                'type': 'local'
            },
            "file_upload_domains": [],
            "web_crawling_allowed_sites": [],
            "cloud": False,
            "jobs": {
                "disable": False
            },
            "tasks": {
                "disable": False
            }
        }
        # endregion

        # region find 'auto' config file, create if not exists
        auto_config_name = 'config.auto.json'
        auto_config_path = self.storage_root_path.joinpath(auto_config_name)
        if not auto_config_path.is_file():
            auto_config_path.write_text('{}')
        self.auto_config_path = auto_config_path
        # endregion

        self.prepare_env_config()

        self.fetch_auto_config()
        self.merge_configs()

        return cls.__instance

    def prepare_env_config(self) -> None:
        """Collect config values from env vars to self._env_config
        """
        self._env_config = {
            'logging': {
                'handlers': {
                    'console': {},
                    'file': {}
                }
            },
            "api": {
                "http": {
                    "server": {}
                }
            },
            'auth': {},
            'paths': {},
            'permanent_storage': {},
            'ml_task_queue': {}
        }

        # region storage root path
        if os.environ.get('MINDSDB_STORAGE_DIR', '') != '':
            self._env_config['paths'] = {
                'root': Path(os.environ['MINDSDB_STORAGE_DIR'])
            }
        # endregion

        # region vars: permanent storage disabled?
        if os.environ.get('MINDSDB_STORAGE_BACKUP_DISABLED', '').lower() in ('1', 'true'):
            self._env_config['permanent_storage'] = {
                'location': 'absent'
            }
        # endregion

        # region vars: ml queue
        if os.environ.get('MINDSDB_ML_QUEUE_TYPE', '').lower() == 'redis':
            self._env_config['ml_task_queue'] = {
                'type': 'redis',
                'host': os.environ.get('MINDSDB_ML_QUEUE_HOST', 'localhost'),
                'port': int(os.environ.get('MINDSDB_ML_QUEUE_PORT', 6379)),
                'db': int(os.environ.get('MINDSDB_ML_QUEUE_DB', 0)),
                'username': os.environ.get('MINDSDB_ML_QUEUE_USERNAME'),
                'password': os.environ.get('MINDSDB_ML_QUEUE_PASSWORD')
            }
        # endregion

        # region vars: username and password
        http_username = os.environ.get('MINDSDB_USERNAME')
        http_password = os.environ.get('MINDSDB_PASSWORD')

        if bool(http_username) != bool(http_password):
            raise ValueError('Both MINDSDB_USERNAME and MINDSDB_PASSWORD must be set together and must be non-empty strings.')

        # If both username and password are set, enable HTTP auth.
        if http_username and http_password:
            self._env_config['auth']['http_auth_enabled'] = True
            self._env_config['auth']['username'] = http_username
            self._env_config['auth']['password'] = http_password
        # endregion

        # region permanent session lifetime
        for env_name in ('MINDSDB_HTTP_PERMANENT_SESSION_LIFETIME', 'FLASK_PERMANENT_SESSION_LIFETIME'):
            env_value = os.environ.get(env_name)
            if isinstance(env_value, str):
                try:
                    permanent_session_lifetime = int(env_value)
                except Exception:
                    raise ValueError(f'Warning: Can\'t cast env var {env_name} value to int: {env_value}')
                self._env_config['auth']['http_permanent_session_lifetime'] = permanent_session_lifetime
                break
        # endregion

        # region logging
        if os.environ.get('MINDSDB_LOG_LEVEL', '') != '':
            self._env_config['logging']['handlers']['console']['level'] = os.environ['MINDSDB_LOG_LEVEL']
            self._env_config['logging']['handlers']['console']['enabled'] = True
        if os.environ.get('MINDSDB_CONSOLE_LOG_LEVEL', '') != '':
            self._env_config['logging']['handlers']['console']['level'] = os.environ['MINDSDB_LOG_LEVEL']
            self._env_config['logging']['handlers']['console']['enabled'] = True
        if os.environ.get('MINDSDB_FILE_LOG_LEVEL', '') != '':
            self._env_config['logging']['handlers']['file']['level'] = os.environ['MINDSDB_FILE_LOG_LEVEL']
            self._env_config['logging']['handlers']['file']['enabled'] = True
        # endregion

        # region server type
        server_type = os.environ.get('MINDSDB_HTTP_SERVER_TYPE', '').lower()
        if server_type == '':
            server_type = os.environ.get('MINDSDB_DEFAULT_SERVER', '').lower()
        if server_type != '':
            if server_type == 'waitress':
                self._env_config['api']['http']['server']['type'] = 'waitress'
                self._default_config['api']['http']['server']['config'] = {}
                self._env_config['api']['http']['server']['config'] = {
                    "threads": 16,
                    "max_request_body_size": (1 << 30) * 10,    # 10GB
                    "inbuf_overflow": (1 << 30) * 10
                }
            elif server_type == 'flask':
                self._env_config['api']['http']['server']['type'] = 'flask'
                self._default_config['api']['http']['server']['config'] = {}
                self._env_config['api']['http']['server']['config'] = {}
            elif server_type == 'gunicorn':
                self._env_config['api']['http']['server']['type'] = 'gunicorn'
                self._default_config['api']['http']['server']['config'] = {}
                self._env_config['api']['http']['server']['config'] = {
                    'workers': min(mp.cpu_count(), 4),
                    'timeout': 600,
                    'reuse_port': True,
                    'preload_app': True,
                    'threads': 4
                }
        # endregion

        if os.environ.get('MINDSDB_DB_CON', '') != '':
            self._env_config['storage_db'] = os.environ['MINDSDB_DB_CON']

    def parse_cmd_args(self) -> None:
        """Collect cmd args to self._cmd_args (accessable as self.cmd_args)
        """
        if self._cmd_args is not None:
            return

        # if it is not mindsdb run, then set args to empty
        if (
            (sys.modules['__main__'].__package__ or '').lower() != 'mindsdb'
            and os.environ.get('MINDSDB_RUNTIME') != "1"
        ):
            self._cmd_args = argparse.Namespace(
                api=None, config=None, install_handlers=None, verbose=False,
                no_studio=False, version=False, ml_task_queue_consumer=None
            )
            return

        parser = argparse.ArgumentParser(description='CL argument for mindsdb server')
        parser.add_argument('--api', type=str, default=None)
        parser.add_argument('--config', type=str, default=None)
        parser.add_argument('--install-handlers', type=str, default=None)
        parser.add_argument('--verbose', action='store_true')
        parser.add_argument('--no_studio', action='store_true')
        parser.add_argument('-v', '--version', action='store_true')
        parser.add_argument('--ml_task_queue_consumer', action='store_true', default=None)
        self._cmd_args = parser.parse_args()

    def fetch_auto_config(self) -> bool:
        """Load dict readed from config.auto.json to `auto_config`.
        Do it only if `auto_config` was not loaded before or config.auto.json been changed.

        Returns:
            bool: True if config was loaded or updated
        """

        if self.auto_config_mtime != self.auto_config_path.stat().st_mtime:
            try:
                self._auto_config = json.loads(self.auto_config_path.read_text())
            except json.JSONDecodeError as e:
                raise ValueError(f"The 'auto' configuration file ({self.auto_config_path}) contains invalid JSON: {e}")
            self.auto_config_mtime = self.auto_config_path.stat().st_mtime
            return True
        return False

    def fetch_user_config(self) -> bool:
        """Read config provided by the user to `user_config`. Do it only if `user_config` was not loaded before.

        Returns:
            bool: True if config was loaded
        """
        if self._user_config is None:
            cmd_args_config = self.cmd_args.config
            if isinstance(cmd_args_config, str):
                self.config_path = cmd_args_config
            elif isinstance(os.environ.get('MINDSDB_CONFIG_PATH'), str):
                self.config_path = os.environ['MINDSDB_CONFIG_PATH']
            if self.config_path == 'absent':
                self.config_path = None
            if isinstance(self.config_path, str):
                self.config_path = Path(self.config_path)
                if not self.config_path.is_file():
                    raise FileNotFoundError(f'The configuration file was not found at the path: {self.config_path}')
                try:
                    self._user_config = json.loads(self.config_path.read_text())
                except json.JSONDecodeError as e:
                    raise ValueError(f'The configuration file ({self.config_path}) contains invalid JSON: {e}')
            else:
                self._user_config = {}
            return True
        return False

    def ensure_auto_config_is_relevant(self) -> None:
        """Check if auto config has not been changed. If changed - reload main config.
        """
        updated = self.fetch_auto_config()
        if updated:
            self.init_config()

    def merge_configs(self) -> None:
        """Merge multiple configs to one.
        """
        new_config = deepcopy(self._default_config)
        _merge_configs(new_config, self._user_config)
        _merge_configs(new_config, self._auto_config)
        _merge_configs(new_config, self._env_config)

        # region create dirs
        for key, value in new_config['paths'].items():
            if isinstance(value, str):
                new_config['paths'][key] = Path(value)
            elif isinstance(value, Path) is False:
                raise ValueError(f"Unexpected path value: {value}")
            create_data_dir(new_config['paths'][key])
        # endregion

        self._config = new_config

    def __getitem__(self, key):
        self.ensure_auto_config_is_relevant()
        return self._config[key]

    def get(self, key, default=None):
        self.ensure_auto_config_is_relevant()
        return self._config.get(key, default)

    def get_all(self):
        self.ensure_auto_config_is_relevant()
        return self._config

    def update(self, data: dict) -> None:
        """Update calues in `auto` config
        """
        self.ensure_auto_config_is_relevant()

        _merge_configs(self._auto_config, data)

        self.auto_config_path.write_text(
            json.dumps(self._auto_config, indent=4)
        )

        self.auto_config_mtime = self.auto_config_path.stat().st_mtime

        self.merge_configs()

    def raise_warnings(self, logger) -> None:
        """Show warnings about config options
        """

        if 'storage_dir' in self._config:
            logger.warning("The 'storage_dir' config option is no longer supported. Use 'paths.root' instead.")

        if 'log' in self._config:
            logger.warning("The 'log' config option is no longer supported. Use 'logging' instead.")

        if os.environ.get('MINDSDB_DEFAULT_SERVER', '') != '':
            logger.warning(
                "Env variable 'MINDSDB_DEFAULT_SERVER' is going to be deprecated soon. "
                "Use 'MINDSDB_HTTP_SERVER_TYPE' instead."
            )

        for env_name in ('MINDSDB_HTTP_SERVER_TYPE', 'MINDSDB_DEFAULT_SERVER'):
            env_value = os.environ.get(env_name, '')
            if env_value.lower() not in ('waitress', 'flask', 'gunicorn'):
                logger.warning(
                    f"The value '{env_value}' of the environment variable {env_name} is not valid. "
                    "It must be one of the following: 'waitress', 'flask', or 'gunicorn'."
                )

    @property
    def cmd_args(self):
        if self._cmd_args is None:
            self.parse_cmd_args()
        return self._cmd_args

    @property
    def paths(self):
        return self._config['paths']

    @property
    def user_config(self):
        return self._user_config

    @property
    def auto_config(self):
        return self._auto_config

    @property
    def env_config(self):
        return self._env_config

    @property
    def is_cloud(self):
        return self._config.get("cloud", False)


config = Config()



"""jobs_active

Revision ID: c67822e96833
Revises: 309db3d07cf4
Create Date: 2023-11-01 15:42:53.249859

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa

# revision identifiers, used by Alembic.
revision = 'c67822e96833'
down_revision = '309db3d07cf4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('active', sa.Boolean(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('active')

    # ### end Alembic commands ###



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    access_token={
        "type": ARG_TYPE.STR,
        "description": " Dropbox Access Token",
        "required": True,
        "label": "Dropbox Access Token",
    },
)

connection_args_example = OrderedDict(
    access_token="ai.L-wqp3eP6r4cSWVklkKAdTNZ3VAuQjWuZMvIs1BzKvZNVW07rKbVNi5HbxvLc9q9D6qSfsf5VTsqYsNPGUkqSJBlpkr88gNboUNuhITmJG9mVw-Olniu4MO3BWVbEIphVxXxxxCd677Y",
)



from mindsdb.integrations.handlers.postgres_handler import Handler as PGHandler


PG_HANDLER_NAME = 'test_handler'
PG_CONNECTION_DATA = {
    "user": "demo_user",
    "password": "demo_password",
    "host": "samples.mindsdb.com",
    "port": "5432",
    "database": "demo"
}


class HandlerControllerMock:
    def __init__(self):
        self.handlers = {
            PG_HANDLER_NAME: PGHandler(
                PG_HANDLER_NAME,
                **{"connection_data": PG_CONNECTION_DATA}
            )
        }

    def get_handler(self, name):
        return self.handlers[name]

    def get(self, name):
        return {
            'id': 0,
            'name': PG_HANDLER_NAME
        }



import requests

from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.libs.api_handler import APIHandler

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response
)

from .serpstack_tables import (
    OrganicResultsTable, ImageResultsTable,
    VideoResultsTable, NewsResultsTable, ShoppingResultsTable
)


logger = log.getLogger(__name__)


class SerpstackHandler(APIHandler):
    """A class for handling connections and interactions with the Serpstack API.
    Attributes:
        api_key (str): API access key for the Serpstack API.
        is_connected (bool): Whether or not the API client is connected to Serpstack.
    """

    name = 'serpstack'

    def __init__(self, name: str = None, **kwargs):
        super().__init__(name)

        connection_data = kwargs.get('connection_data', {})
        self.connection_data = connection_data

        self.access_key = None
        self.base_url = None
        self.is_connected = False

        if 'access_key' in self.connection_data:
            self.access_key = self.connection_data['access_key']

        # register tables
        organic_results = OrganicResultsTable(self)
        self._register_table('organic_results', organic_results)

        image_results = ImageResultsTable(self)
        self._register_table('image_results', image_results)

        video_results = VideoResultsTable(self)
        self._register_table('video_results', video_results)

        news_results = NewsResultsTable(self)
        self._register_table('news_results', news_results)

        shopping_results = ShoppingResultsTable(self)
        self._register_table('shopping_results', shopping_results)

    def connect(self):
        """ Sets up connection and returns the base URL to be used"""

        if self.is_connected:
            return self.base_url

        if not self.access_key:
            logger.error("No access key provided for Serpstack API")
            return None

        try:
            url = f"https://api.serpstack.com/search?access_key={self.access_key}"
            api_request = requests.get(url)
            api_response = api_request.json()
            # error 105 means that user is on a free plan
            if api_response['error']['code'] == 105:
                self.base_url = "http://api.serpstack.com/search"
            # error 310 means that missing search query, which means that user can use https
            elif api_response['error']['code'] == 310:
                self.base_url = "https://api.serpstack.com/search"
            # any other error suggests issues with the account
            else:
                logger.error(f"Failed to connect to Serpstack API: {api_response['error']['info']}")
                return None

            self.is_connected = True
            return self.base_url

        except Exception as e:
            logger.error(f"Failed to connect to Serpstack API: {str(e)}")
            return None

    def check_connection(self) -> StatusResponse:
        """ Checks connection to Serpstack API"""
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
            response.copy_storage = True
        except Exception as e:
            response.error_message = (
                f"Failed to connect to Serpstack API: {str(e)}"
            )
            logger.error(response.error_message)
            response.success = False

        self.is_connected = response.success

        return response

    def native_query(self, query: str = None) -> Response:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query)
        return self.query(ast)



# Default settings for markdown header splitting
DEFAULT_MARKDOWN_HEADERS = [
    ("#", "Header 1"),
    ("##", "Header 2"),
    ("###", "Header 3"),
]

# Limits for web crawling
DEFAULT_CRAWL_DEPTH = 1
DEFAULT_WEB_CRAWL_LIMIT = 100
DEFAULT_WEB_FILTERS = []

DEFAULT_CONTEXT_DOCUMENT_LIMIT = 50



import json
import requests


class LumaClient:
    def __init__(self, api_key):
        self.auth_token = api_key
        self.luma_base_endpoint = "https://api.lu.ma/"
        self.validate_api_key()

    def make_request(self, url, method='GET', payload=None):
        if method not in ['GET', 'POST']:
            raise ValueError('Invalid HTTP request method')
        headers = {"accept": "application/json", "content-type": "application/json"}
        if self.auth_token:
            headers['x-luma-api-key'] = self.auth_token
        request_method = getattr(requests, method.lower())
        resp = request_method(url, json=payload, headers=headers)
        return {"content": resp.json(), "code": resp.status_code}

    def validate_api_key(self):
        url = f'{self.luma_base_endpoint}public/v1/user/get-self'
        content = self.make_request(url)
        if content['code'] != 200:
            raise Exception("User Authentication failed - " + json.dumps(content["content"]))
        return content

    def create_event(self, data):
        url = f'{self.luma_base_endpoint}public/v1/event/create'
        content = self.make_request(url, method="POST", payload=data)
        if content['code'] != 200:
            raise Exception("Create failed - " + json.dumps(content["content"]))
        return content

    def get_event(self, event_api_id):
        url = f'{self.luma_base_endpoint}public/v1/event/get?api_id={event_api_id}'
        content = self.make_request(url)
        if content['code'] != 200:
            raise Exception("Get event failed - " + json.dumps(content["content"]))
        return content

    def list_events(self):
        url = f'{self.luma_base_endpoint}public/v1/calendar/list-events?series_mode=sessions'
        content = self.make_request(url)
        if content['code'] != 200:
            raise Exception("Get event failed - " + json.dumps(content["content"]))
        return content



import unittest
from mindsdb.integrations.handlers.openstreetmap_handler.openstreetmap_handler import OpenStreetMapHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class OpenStreetMapHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.handler = OpenStreetMapHandler(name='test_handler', connection_data={})

    def test_0_connect(self):
        assert self.handler.connect()

    def test_1_check_connection(self):
        assert self.handler.check_connection()

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_3_select_query(self):
        query = "SELECT * FROM openstreetmap_datasource.nodes WHERE id = 1;"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_native_query(self):
        query = "SELECT * FROM openstreetmap_datasource.nodes WHERE area = 'New Delhi';"
        response = self.handler.native_query(query)
        assert response.type is RESPONSE_TYPE.ERROR

    def test_5_disconnect(self):
        assert self.handler.disconnect()


if __name__ == '__main__':
    unittest.main()



import copy

from mindsdb_sql_parser import ast
from mindsdb_sql_parser.ast import (
    Select, Identifier, Join, Star, BinaryOperation, Constant, Union, CreateTable,
    Function, Insert, Except, Intersect, Update, NativeQuery, Parameter, Delete
)

from mindsdb.api.executor.planner.exceptions import PlanningException
from mindsdb.api.executor.planner import utils
from mindsdb.api.executor.planner.query_plan import QueryPlan
from mindsdb.api.executor.planner.steps import (
    FetchDataframeStep, ProjectStep, ApplyPredictorStep,
    ApplyPredictorRowStep, UnionStep, GetPredictorColumns, SaveToTable,
    InsertToTable, UpdateToTable, SubSelectStep,
    DeleteStep, DataStep, CreateTableStep
)
from mindsdb.api.executor.planner.utils import (
    disambiguate_predictor_column_identifier,
    get_deepest_select,
    recursively_extract_column_values,
    query_traversal, filters_to_bin_op
)
from mindsdb.api.executor.planner.plan_join import PlanJoin
from mindsdb.api.executor.planner.query_prepare import PreparedStatementPlanner


class QueryPlanner:

    def __init__(self,
                 query=None,
                 integrations: list = None,
                 predictor_namespace=None,
                 predictor_metadata: list = None,
                 default_namespace: str = None):
        self.query = query
        self.plan = QueryPlan()

        _projects = set()
        self.integrations = {}
        if integrations is not None:
            for integration in integrations:
                if isinstance(integration, dict):
                    integration_name = integration['name'].lower()
                    # it is project of system database
                    if integration['type'] != 'data':
                        _projects.add(integration_name)
                        continue
                else:
                    integration_name = integration.lower()
                    integration = {'name': integration}
                self.integrations[integration_name] = integration

        # allow to select from mindsdb namespace
        _projects.add('mindsdb')

        self.default_namespace = default_namespace

        # legacy parameter
        self.predictor_namespace = predictor_namespace.lower() if predictor_namespace else 'mindsdb'

        # map for lower names of predictors

        self.predictor_info = {}
        if isinstance(predictor_metadata, list):
            # convert to dict
            for predictor in predictor_metadata:
                if 'integration_name' in predictor:
                    integration_name = predictor['integration_name']
                else:
                    integration_name = self.predictor_namespace
                    predictor['integration_name'] = integration_name
                idx = f'{integration_name}.{predictor["name"]}'.lower()
                self.predictor_info[idx] = predictor
                _projects.add(integration_name.lower())
        elif isinstance(predictor_metadata, dict):
            # legacy behaviour
            for name, predictor in predictor_metadata.items():
                if '.' not in name:
                    if 'integration_name' in predictor:
                        integration_name = predictor['integration_name']
                    else:
                        integration_name = self.predictor_namespace
                        predictor['integration_name'] = integration_name
                    name = f'{integration_name}.{name}'.lower()
                    _projects.add(integration_name.lower())

                self.predictor_info[name] = predictor

        self.projects = list(_projects)
        self.databases = list(self.integrations.keys()) + self.projects

        self.statement = None

        self.cte_results = {}

    def is_predictor(self, identifier):
        if not isinstance(identifier, Identifier):
            return False
        return self.get_predictor(identifier) is not None

    def get_predictor(self, identifier):
        name_parts = list(identifier.parts)

        version = None
        if len(name_parts) > 1 and name_parts[-1].isdigit():
            # last part is version
            version = name_parts[-1]
            name_parts = name_parts[:-1]

        name = name_parts[-1]

        namespace = None
        if len(name_parts) > 1:
            namespace = name_parts[-2]
        else:
            if self.default_namespace is not None:
                namespace = self.default_namespace

        idx_ar = [name]
        if namespace is not None:
            idx_ar.insert(0, namespace)

        idx = '.'.join(idx_ar).lower()
        info = self.predictor_info.get(idx)
        if info is not None:
            info['version'] = version
            info['name'] = name
        return info

    def prepare_integration_select(self, database, query):
        # replacement for 'utils.recursively_disambiguate_*' functions from utils
        #   main purpose: make tests working (don't change planner outputs)
        # can be removed in future (with adapting the tests) except 'cut integration part' block

        def _prepare_integration_select(node, is_table, is_target, parent_query, **kwargs):
            if not isinstance(node, Identifier):
                return

            # cut integration part
            if len(node.parts) > 1 and node.parts[0].lower() == database:
                node.parts.pop(0)

            if not hasattr(parent_query, 'from_table'):
                return

            table = parent_query.from_table
            if not is_table:
                # add table name or alias for identifiers
                if isinstance(table, Join):
                    # skip for join
                    return

                # keep column name for target
                if is_target:
                    if node.alias is None:
                        last_part = node.parts[-1]
                        if isinstance(last_part, str):
                            node.alias = Identifier(parts=[node.parts[-1]])

        query_traversal(query, _prepare_integration_select)

    def get_integration_select_step(self, select):
        if isinstance(select.from_table, NativeQuery):
            integration_name = select.from_table.integration.parts[-1]
        else:
            integration_name, table = self.resolve_database_table(select.from_table)

            # is it CTE?
            table_name = table_alias = table.parts[-1]
            if table.alias is not None:
                table_alias = table.alias.parts[-1]

            if integration_name == self.default_namespace and table_name in self.cte_results:
                select.from_table = None
                return SubSelectStep(select, self.cte_results[table_name], table_name=table_alias)

        fetch_df_select = copy.deepcopy(select)
        self.prepare_integration_select(integration_name, fetch_df_select)

        # remove predictor params
        if fetch_df_select.using is not None:
            fetch_df_select.using = None

        return FetchDataframeStep(integration=integration_name, query=fetch_df_select)

    def plan_integration_select(self, select):
        """Plan for a select query that can be fully executed in an integration"""

        return self.plan.add_step(self.get_integration_select_step(select))

    def resolve_database_table(self, node: Identifier):
        # resolves integration name and table name

        parts = node.parts.copy()
        alias = None
        if node.alias is not None:
            alias = node.alias.copy()

        database = self.default_namespace

        if len(parts) > 1:
            if parts[0].lower() in self.databases:
                database = parts.pop(0).lower()

        if database is None:
            raise PlanningException(f'Integration not found for: {node}')

        return database, Identifier(parts=parts, alias=alias)

    def get_query_info(self, query):
        # get all predictors
        mdb_entities = []
        predictors = []
        user_functions = []
        # projects = set()
        integrations = set()

        def find_objects(node, is_table, **kwargs):
            if isinstance(node, Function):
                if node.namespace is not None or node.op.lower() in ('llm',):
                    user_functions.append(node)

            if is_table:
                if isinstance(node, ast.Identifier):
                    integration, _ = self.resolve_database_table(node)

                    if self.is_predictor(node):
                        predictors.append(node)

                    if integration in self.projects:
                        # it is project
                        mdb_entities.append(node)

                    elif integration is not None:
                        integrations.add(integration)
                if isinstance(node, ast.NativeQuery) or isinstance(node, ast.Data):
                    mdb_entities.append(node)

        query_traversal(query, find_objects)

        # cte names are not mdb objects
        if isinstance(query, Select) and query.cte:
            cte_names = [
                cte.name.parts[-1]
                for cte in query.cte
            ]
            mdb_entities = [
                item
                for item in mdb_entities
                if '.'.join(item.parts) not in cte_names
            ]

        return {
            'mdb_entities': mdb_entities,
            'integrations': integrations,
            'predictors': predictors,
            'user_functions': user_functions
        }

    def get_nested_selects_plan_fnc(self, main_integration, force=False):
        # returns function for traversal over query and inject fetch data query instead of subselects
        def find_selects(node, **kwargs):
            if isinstance(node, Select):
                query_info2 = self.get_query_info(node)
                if force or (
                        len(query_info2['integrations']) > 1
                        or main_integration not in query_info2['integrations']
                        or len(query_info2['mdb_entities']) > 0
                ):
                    # need to execute in planner

                    node.parentheses = False
                    last_step = self.plan_select(node)

                    node2 = Parameter(last_step.result)

                    return node2

        return find_selects

    def plan_select_identifier(self, query):
        # query_info = self.get_query_info(query)
        #
        # if len(query_info['integrations']) == 0 and len(query_info['predictors']) >= 1:
        #     # select from predictor
        #     return self.plan_select_from_predictor(query)
        # elif (
        #     len(query_info['integrations']) == 1
        #     and len(query_info['mdb_entities']) == 0
        #     and len(query_info['user_functions']) == 0
        # ):
        #
        #     int_name = list(query_info['integrations'])[0]
        #     if self.integrations.get(int_name, {}).get('class_type') != 'api':
        #         # one integration without predictors, send all query to integration
        #         return self.plan_integration_select(query)

        # find subselects
        main_integration, _ = self.resolve_database_table(query.from_table)
        is_api_db = self.integrations.get(main_integration, {}).get('class_type') == 'api'

        find_selects = self.get_nested_selects_plan_fnc(main_integration, force=is_api_db)
        query.targets = query_traversal(query.targets, find_selects)
        query_traversal(query.where, find_selects)

        # get info of updated query
        query_info = self.get_query_info(query)

        if len(query_info['predictors']) >= 1:
            # select from predictor
            return self.plan_select_from_predictor(query)
        elif is_api_db:
            return self.plan_api_db_select(query)
        elif len(query_info['user_functions']) > 0:
            return self.plan_integration_select_with_functions(query)
        else:
            # fallback to integration
            return self.plan_integration_select(query)

    def plan_integration_select_with_functions(self, query):
        # UDF can't be aggregate function: it means we have to do aggregation after function execution
        # - remove targets from query
        # - add subselect with targets

        # replace functions in conditions

        query2 = query.copy()

        skipped_conditions = []

        def replace_functions(node, **kwargs):
            if not isinstance(node, BinaryOperation):
                return

            arg1, arg2 = node.args
            if not isinstance(arg1, Function):
                arg1 = arg2
            if not isinstance(arg1, Function):
                return

            # user defined
            if arg1.namespace is not None:
                # clear
                skipped_conditions.append(node)
                node.args = [Constant(0), Constant(0)]
                node.op = '='

        query_traversal(query2.where, replace_functions)

        query2.targets = [Star()]

        # don't do aggregate
        query2.having = None

        if query.group_by is not None:
            # if aggregation exists, do order and limit in subquery
            query2.group_by = None
            query2.order_by = None
            query2.limit = None
        else:
            query.order_by = None
            query.limit = None

        # if all conditions were executed - clear it
        if len(skipped_conditions) == 0:
            query.where = None

        prev_step = self.plan_integration_select(query2)

        return self.plan_sub_select(query, prev_step)

    def plan_api_db_select(self, query):
        # split to select from api database
        #     keep only limit and where
        #     the rest goes to outer select
        query2 = Select(
            targets=query.targets,
            from_table=query.from_table,
            where=query.where,
            order_by=query.order_by,
            limit=query.limit,
        )
        prev_step = self.plan_integration_select(query2)

        # clear limit and where
        query.limit = None
        query.where = None
        return self.plan_sub_select(query, prev_step)

    def plan_nested_select(self, select):

        # query_info = self.get_query_info(select)
        # # get all predictors
        #
        # if (
        #     len(query_info['mdb_entities']) == 0
        #     and len(query_info['integrations']) == 1
        #     and len(query_info['user_functions']) == 0
        #     and 'files' not in query_info['integrations']
        #     and 'views' not in query_info['integrations']
        # ):
        #     int_name = list(query_info['integrations'])[0]
        #     if self.integrations.get(int_name, {}).get('class_type') != 'api':
        #
        #         # if no predictor inside = run as is
        #         return self.plan_integration_nested_select(select, int_name)

        return self.plan_mdb_nested_select(select)

    def plan_integration_nested_select(self, select, integration_name):
        fetch_df_select = copy.deepcopy(select)
        deepest_select = get_deepest_select(fetch_df_select)
        self.prepare_integration_select(integration_name, deepest_select)
        return self.plan.add_step(FetchDataframeStep(integration=integration_name, query=fetch_df_select))

    def plan_mdb_nested_select(self, select):
        # plan nested select

        select2 = copy.deepcopy(select.from_table)
        select2.parentheses = False
        select2.alias = None
        self.plan_select(select2)
        last_step = self.plan.steps[-1]

        return self.plan_sub_select(select, last_step)

    def get_predictor_namespace_and_name_from_identifier(self, identifier):
        new_identifier = copy.deepcopy(identifier)

        info = self.get_predictor(identifier)
        namespace = info['integration_name']

        parts = [namespace, info['name']]
        if info['version'] is not None:
            parts.append(info['version'])
        new_identifier.parts = parts

        return namespace, new_identifier

    def plan_select_from_predictor(self, select):
        predictor_namespace, predictor = self.get_predictor_namespace_and_name_from_identifier(select.from_table)

        if select.where == BinaryOperation('=', args=[Constant(1), Constant(0)]):
            # Hardcoded mysql way of getting predictor columns
            predictor_identifier = utils.get_predictor_name_identifier(predictor)
            predictor_step = self.plan.add_step(
                GetPredictorColumns(
                    namespace=predictor_namespace,
                    predictor=predictor_identifier
                )
            )
        else:
            new_query_targets = []
            for target in select.targets:
                if isinstance(target, Identifier):
                    new_query_targets.append(
                        disambiguate_predictor_column_identifier(target, predictor))
                elif type(target) in (Star, Constant, Function):
                    new_query_targets.append(target)
                else:
                    raise PlanningException(f'Unknown select target {type(target)}')

            if select.group_by or select.having:
                raise PlanningException('Unsupported operation when querying predictor. Only WHERE is allowed and required.')

            row_dict = {}
            where_clause = select.where
            if not where_clause:
                raise PlanningException('WHERE clause required when selecting from predictor')

            predictor_identifier = utils.get_predictor_name_identifier(predictor)
            recursively_extract_column_values(where_clause, row_dict, predictor_identifier)

            params = None
            if select.using is not None:
                params = select.using
            predictor_step = self.plan.add_step(
                ApplyPredictorRowStep(
                    namespace=predictor_namespace,
                    predictor=predictor_identifier,
                    row_dict=row_dict,
                    params=params
                )
            )
        project_step = self.plan_project(select, predictor_step.result)
        return project_step

    def plan_predictor(self, query, table, predictor_namespace, predictor):
        int_select = copy.deepcopy(query)
        int_select.targets = [Star()]  # TODO why not query.targets?
        int_select.from_table = table

        predictor_alias = None
        if predictor.alias is not None:
            predictor_alias = predictor.alias.parts[0]

        params = {}
        if query.using is not None:
            params = query.using

        binary_ops = []
        table_filters = []
        model_filters = []

        def split_filters(node, **kwargs):
            # split conditions between model and table

            if isinstance(node, BinaryOperation):
                op = node.op.lower()

                binary_ops.append(op)

                if op in ['and', 'or']:
                    return

                arg1, arg2 = node.args
                if not isinstance(arg1, Identifier):
                    arg1, arg2 = arg2, arg1

                if isinstance(arg1, Identifier) and isinstance(arg2, (Constant, Parameter)) and len(arg1.parts) > 1:
                    model = Identifier(parts=arg1.parts[:-1])

                    if (
                        self.is_predictor(model)
                        or (
                            len(model.parts) == 1 and model.parts[0] == predictor_alias
                        )
                    ):
                        model_filters.append(node)
                        return
                table_filters.append(node)

        # find subselects
        main_integration, _ = self.resolve_database_table(table)
        find_selects = self.get_nested_selects_plan_fnc(main_integration, force=True)
        query_traversal(int_select.where, find_selects)

        # split conditions
        query_traversal(int_select.where, split_filters)

        if len(model_filters) > 0 and 'or' not in binary_ops:
            int_select.where = filters_to_bin_op(table_filters)

        integration_select_step = self.plan_integration_select(int_select)

        predictor_identifier = utils.get_predictor_name_identifier(predictor)

        if len(params) == 0:
            params = None

        row_dict = None
        if model_filters:
            row_dict = {}
            for el in model_filters:
                if isinstance(el.args[0], Identifier) and el.op == '=':
                    if isinstance(el.args[1], (Constant, Parameter)):
                        row_dict[el.args[0].parts[-1]] = el.args[1].value

        last_step = self.plan.add_step(ApplyPredictorStep(
            namespace=predictor_namespace,
            dataframe=integration_select_step.result,
            predictor=predictor_identifier,
            params=params,
            row_dict=row_dict
        ))

        return {
            'predictor': last_step,
            'data': integration_select_step,
        }

    # def plan_group(self, query, last_step):
    #     # ! is not using yet
    #
    #     # check group
    #     funcs = []
    #     for t in query.targets:
    #         if isinstance(t, Function):
    #             funcs.append(t.op.lower())
    #     agg_funcs = ['sum', 'min', 'max', 'avg', 'count', 'std']
    #
    #     if (
    #             query.having is not None
    #             or query.group_by is not None
    #             or set(agg_funcs) & set(funcs)
    #     ):
    #         # is aggregate
    #         group_by_targets = []
    #         for t in query.targets:
    #             target_copy = copy.deepcopy(t)
    #             group_by_targets.append(target_copy)
    #         # last_step = self.plan.steps[-1]
    #         return GroupByStep(dataframe=last_step.result, columns=query.group_by, targets=group_by_targets)

    def plan_project(self, query, dataframe, ignore_doubles=False):
        out_identifiers = []

        if len(query.targets) == 1 and isinstance(query.targets[0], Star):
            last_step = self.plan.steps[-1]
            return last_step

        for target in query.targets:
            if isinstance(target, Identifier) \
                    or isinstance(target, Star) \
                    or isinstance(target, Function) \
                    or isinstance(target, Constant) \
                    or isinstance(target, BinaryOperation):
                out_identifiers.append(target)
            else:
                new_identifier = Identifier(str(target.to_string(alias=False)), alias=target.alias)
                out_identifiers.append(new_identifier)
        return self.plan.add_step(ProjectStep(dataframe=dataframe, columns=out_identifiers, ignore_doubles=ignore_doubles))

    def plan_create_table(self, query: CreateTable):
        if query.from_select is None:
            if query.columns is not None:
                self.plan.add_step(CreateTableStep(
                    table=query.name,
                    columns=query.columns,
                    is_replace=query.is_replace,
                ))
                return

            raise PlanningException(f'Not implemented "create table": {query.to_string()}')

        integration_name = query.name.parts[0]

        last_step = self.plan_select(query.from_select, integration=integration_name)

        # create table step
        self.plan.add_step(SaveToTable(
            table=query.name,
            dataframe=last_step,
            is_replace=query.is_replace,
        ))

    def plan_insert(self, query):
        table = query.table
        if query.from_select is not None:
            integration_name = query.table.parts[0]

            # plan sub-select first
            last_step = self.plan_select(query.from_select, integration=integration_name)

            self.plan.add_step(InsertToTable(
                table=table,
                dataframe=last_step,
            ))
        else:
            self.plan.add_step(InsertToTable(
                table=table,
                query=query,
            ))

    def plan_update(self, query):
        last_step = None
        if query.from_select is not None:
            integration_name = query.table.parts[0]
            last_step = self.plan_select(query.from_select, integration=integration_name)

        # plan sub-select first
        update_command = copy.deepcopy(query)
        # clear subselect
        update_command.from_select = None

        table = query.table
        self.plan.add_step(UpdateToTable(
            table=table,
            dataframe=last_step,
            update_command=update_command
        ))

    def plan_delete(self, query: Delete):

        # find subselects
        main_integration, _ = self.resolve_database_table(query.table)

        is_api_db = self.integrations.get(main_integration, {}).get('class_type') == 'api'

        find_selects = self.get_nested_selects_plan_fnc(main_integration, force=is_api_db)
        query_traversal(query.where, find_selects)

        self.prepare_integration_select(main_integration, query.where)

        return self.plan.add_step(DeleteStep(
            table=query.table,
            where=query.where
        ))

    def plan_cte(self, query):

        for cte in query.cte:
            step = self.plan_select(cte.query)
            name = cte.name.parts[-1]
            self.cte_results[name] = step.result

    def check_single_integration(self, query):
        query_info = self.get_query_info(query)

        # can we send all query to integration?

        # one integration and not mindsdb objects in query
        if (
                len(query_info['mdb_entities']) == 0
                and len(query_info['integrations']) == 1
                and 'files' not in query_info['integrations']
                and 'views' not in query_info['integrations']
                and len(query_info['user_functions']) == 0
        ):

            int_name = list(query_info['integrations'])[0]
            # if is sql database
            if self.integrations.get(int_name, {}).get('class_type') != 'api':

                # send to this integration
                self.prepare_integration_select(int_name, query)

                last_step = self.plan.add_step(FetchDataframeStep(integration=int_name, query=query))
                return last_step

    def plan_select(self, query, integration=None):

        if isinstance(query, (Union, Except, Intersect)):
            return self.plan_union(query, integration=integration)

        if query.cte is not None:
            self.plan_cte(query)

        from_table = query.from_table

        if isinstance(from_table, Identifier):
            return self.plan_select_identifier(query)
        elif isinstance(from_table, Select):
            return self.plan_nested_select(query)
        elif isinstance(from_table, Join):
            plan_join = PlanJoin(self)
            return plan_join.plan(query, integration)
        elif isinstance(from_table, NativeQuery):
            integration = from_table.integration.parts[0].lower()
            step = FetchDataframeStep(integration=integration, raw_query=from_table.query)
            last_step = self.plan.add_step(step)
            return self.plan_sub_select(query, last_step)

        elif isinstance(from_table, ast.Data):
            step = DataStep(from_table.data)
            last_step = self.plan.add_step(step)
            return self.plan_sub_select(query, last_step, add_absent_cols=True)

        else:
            raise PlanningException(f'Unsupported from_table {type(from_table)}')

    def plan_sub_select(self, query, prev_step, add_absent_cols=False):
        if (
            query.group_by is not None
            or query.order_by is not None
            or query.having is not None
            or query.distinct is True
            or query.where is not None
            or query.limit is not None
            or query.offset is not None
            or len(query.targets) != 1
            or not isinstance(query.targets[0], Star)
        ):
            if query.from_table.alias is not None:
                table_name = query.from_table.alias.parts[-1]
            elif isinstance(query.from_table, Identifier):
                table_name = query.from_table.parts[-1]
            else:
                table_name = None

            query2 = copy.deepcopy(query)
            query2.from_table = None
            sup_select = SubSelectStep(query2, prev_step.result, table_name=table_name, add_absent_cols=add_absent_cols)
            self.plan.add_step(sup_select)
            return sup_select
        return prev_step

    def plan_union(self, query, integration=None):
        step1 = self.plan_select(query.left, integration=integration)
        step2 = self.plan_select(query.right, integration=integration)
        operation = 'union'
        if isinstance(query, Except):
            operation = 'except'
        elif isinstance(query, Intersect):
            operation = 'intersect'

        return self.plan.add_step(UnionStep(left=step1.result, right=step2.result, unique=query.unique, operation=operation))

    # method for compatibility
    def from_query(self, query=None):
        self.plan = QueryPlan()

        if query is None:
            query = self.query

        if isinstance(query, (Select, Union, Except, Intersect)):
            if self.check_single_integration(query):
                return self.plan
            self.plan_select(query)
        elif isinstance(query, CreateTable):
            self.plan_create_table(query)
        elif isinstance(query, Insert):
            self.plan_insert(query)
        elif isinstance(query, Update):
            self.plan_update(query)
        elif isinstance(query, Delete):
            self.plan_delete(query)
        else:
            raise PlanningException(f'Unsupported query type {type(query)}')

        return self.plan

    def prepare_steps(self, query):
        statement_planner = PreparedStatementPlanner(self)

        # return generator
        return statement_planner.prepare_steps(query)

    def execute_steps(self, params=None):
        statement_planner = PreparedStatementPlanner(self)

        # return generator
        return statement_planner.execute_steps(params)

    # def fetch(self, row_count):
    #     statement_planner = PreparedStatementPlanner(self)
    #     return statement_planner.fetch(row_count)
    #
    # def close(self):
    #     statement_planner = PreparedStatementPlanner(self)
    #     return statement_planner.close()

    def get_statement_info(self):
        statement_planner = PreparedStatementPlanner(self)

        return statement_planner.get_statement_info()



from mindsdb_sql_parser import ast
import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

class CatalogSearchTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:

        """The select method implements the mappings from the ast.Select and calls the actual API through the call_mendeley_api.
        Firstly, it is used to put the parameters specified in the query in a dictionary, which is then used when calling the method call_mendeley_api.
        If no conditions are specified, an error is raised since the search cannot be conducted.

        Args:
            query (ast.Select): query used to specify the wanted results 
        Returns:
            result (DataFrame): the result of the query
        """
       
        conditions = extract_comparison_conditions(query.where)
        
        params = {} 

        #Since there are three different types of search, and each of them takes different parameters, we use the parameters that lead
        # to the most specific results. For example, in the case of the user specifying the title and the doi of a document, priority is given to
        # the doi.

        if query.limit is not None:
            params['limit'] = query.limit.value
        else:
            params['limit'] = 30

        for op,arg1,arg2 in conditions:

            if arg1 in ['arxiv', 'doi', 'isbn', 'issn', 'pmid', 'scopus', 'filehash']:

                if op != '=':
                    raise NotImplementedError
                params[arg1] = arg2

                result = self.handler.call_mendeley_api(
                method_name='identifier_search',
                params=params)

                break

            elif arg1 == 'id':
                if op != '=':
                    raise NotImplementedError
                params['id'] = arg2

                result = self.handler.call_mendeley_api(
                method_name='get',
                params=params)

                break

            elif "title" or "author" or"source" or "abstract" or "min_year" or "max_year" or "open_access" or "view" in conditions:

                if arg1 == 'title':
                    if op != '=':
                        raise NotImplementedError
                    params['title'] = arg2
                
                elif arg1 == 'author':
                    if op != '=':
                        raise NotImplementedError
                    params['author'] = arg2
                
                elif arg1 == 'source':
                    if op != '=':
                        raise NotImplementedError
                    params['source'] = arg2
                
                elif arg1 == 'abstract':
                    if op != '=':
                        raise NotImplementedError
                    params['abstract'] = arg2
                
                elif arg1 == 'min_year':
                    params['min_year'] = arg2
                
                elif arg1 == 'max_year':
                    params['max_year'] = arg2
                
                elif arg1 == 'open_access':
                    if op != '=':
                        raise NotImplementedError
                    params['open_access'] = arg2
                
            
                result = self.handler.call_mendeley_api(
                method_name='advanced_search',
                params=params)
        
        if conditions==[]:
            raise ValueError('Please give input for the search to be conducted.')
        

        columns = []

        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")
        

        return result[columns]
    
    def get_columns(self):
        """ get_columns method returns the columns returned by the API"""
        return [

            'title',
            'type',
            'source',
            'year',
            'pmid',
            'sgr',
            'issn',
            'scopus',
            'doi',
            'pui',
            'authors',
            'keywords',
            'link',
            'id'
        ]

        


from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    db_file={
        'type': ARG_TYPE.STR,
        'description': 'The database file where the data will be stored.'
    }
)

connection_args_example = OrderedDict(
    db_file='C:\\Users\\minurap\\Documents\\example_db.accdb'
)



from typing import Any, List
from langchain_core.embeddings import Embeddings
from openai import AsyncOpenAI
import asyncio


class VLLMEmbeddings(Embeddings):
    """VLLMEmbeddings uses a VLLM server to generate embeddings."""

    def __init__(
        self,
        openai_api_base: str,
        model: str,
        batch_size: int = 32,
        **kwargs: Any,
    ):
        """Initialize the embeddings class.

        Args:
            openai_api_base: Base URL for the VLLM server
            model: Model name/path to use for embeddings
            batch_size: Batch size for generating embeddings
        """
        super().__init__()
        self.model = model
        self.batch_size = batch_size
        self.is_nomic = "nomic-embed-text" in model.lower()

        # Initialize OpenAI client
        openai_kwargs = kwargs.copy()
        if "input_columns" in openai_kwargs:
            del openai_kwargs["input_columns"]

        self.client = AsyncOpenAI(
            api_key="EMPTY",  # vLLM doesn't need an API key
            base_url=openai_api_base,
            **openai_kwargs,
        )

    def _format_text(self, text: str, is_query: bool = False) -> str:
        """
        Format text according to nomic-embed requirements if using nomic model.
        e.g. see here for more details: https://huggingface.co/nomic-ai/nomic-embed-text-v1.5#task-instruction-prefixes
        """

        if not self.is_nomic:
            return text
        prefix = "search_query: " if is_query else "search_document: "
        return prefix + text

    def _get_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Get embeddings for a batch of texts."""

        async def await_openai_call(batch):
            return await self.client.embeddings.create(model=self.model, input=batch)

        embeddings = []
        embedding_coroutines = []
        chunk_start_indices = range(0, len(texts), self.batch_size)
        for i in chunk_start_indices:

            batch = texts[i: i + self.batch_size]
            embedding_coroutines.append(await_openai_call(batch))

            # if at max-concurrency, then run with gather
            if len(embedding_coroutines) == 512 or len(embedding_coroutines) == len(
                chunk_start_indices
            ):

                openai_responses = []

                async def gather_coroutines(openai_responses):
                    # define a function to gather and save responses.
                    intermediate = await asyncio.gather(*embedding_coroutines)
                    openai_responses.extend(intermediate)

                # run asynchronously
                asyncio.run(gather_coroutines(openai_responses))

                # extract embeddings from responses
                for response in openai_responses:
                    embeddings.extend([data.embedding for data in response.data])

                # reset the embedding_coroutines list
                embedding_coroutines = []

        return embeddings

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed a list of documents using vLLM.

        Args:
            texts: List of documents to embed

        Returns:
            List of embeddings, one for each document
        """
        formatted_texts = [self._format_text(text) for text in texts]
        return self._get_embeddings(formatted_texts)

    def embed_query(self, text: str) -> List[float]:
        """Embed a single query text using vLLM.

        Args:
            text: Query text to embed

        Returns:
            Query embedding
        """
        formatted_text = self._format_text(text, is_query=True)
        return self._get_embeddings([formatted_text])[0]



"""store llm data

Revision ID: cbedc4968d5d
Revises: 2958416fbe75
Create Date: 2024-06-06 13:59:45.158089

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = 'cbedc4968d5d'
down_revision = '2958416fbe75'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'llm_data',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('input', sa.String(), nullable=False),
        sa.Column('output', sa.String(), nullable=False),
        sa.Column('model_id', sa.String(), nullable=False),
        sa.Column('created_at', sa.DateTime(), server_default=sa.func.now()),
        sa.Column('updated_at', sa.DateTime(), server_default=sa.func.now(), onupdate=sa.func.now()),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('llm_data')



"""Factory functions for creating retrievers."""

from mindsdb.integrations.utilities.rag.settings import RAGPipelineModel, RetrieverType
from mindsdb.integrations.utilities.rag.vector_store import VectorStoreOperator
from mindsdb.integrations.utilities.rag.retrievers.auto_retriever import AutoRetriever
from mindsdb.integrations.utilities.rag.retrievers.sql_retriever import SQLRetriever


def create_vector_store_retriever(config: RAGPipelineModel):
    """Create a vector store retriever."""
    if getattr(config.vector_store, '_mock_return_value', None) is not None:
        # If vector_store is mocked, return a simple mock retriever for testing
        from unittest.mock import MagicMock
        mock_retriever = MagicMock()
        mock_retriever._get_relevant_documents.return_value = [
            {"page_content": "The Wright brothers invented the airplane."}
        ]
        return mock_retriever

    vector_store_operator = VectorStoreOperator(
        vector_store=config.vector_store,
        documents=config.documents,
        embedding_model=config.embedding_model,
        vector_store_config=config.vector_store_config
    )
    return vector_store_operator.vector_store.as_retriever()


def create_auto_retriever(config: RAGPipelineModel):
    """Create an auto retriever."""
    return AutoRetriever(
        vector_store=config.vector_store,
        documents=config.documents,
        embedding_model=config.embedding_model
    )


def create_sql_retriever(config: RAGPipelineModel):
    """Create a SQL retriever."""
    return SQLRetriever(
        sql_source=config.sql_source,
        llm=config.llm
    )


def create_retriever(config: RAGPipelineModel, retriever_type: RetrieverType = None):
    """Create a retriever based on type."""
    retriever_type = retriever_type or config.retriever_type

    if retriever_type == RetrieverType.VECTOR_STORE:
        return create_vector_store_retriever(config)
    elif retriever_type == RetrieverType.AUTO:
        return create_auto_retriever(config)
    elif retriever_type == RetrieverType.SQL:
        return create_sql_retriever(config)
    else:
        raise ValueError(f"Unsupported retriever type: {retriever_type}")



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        'type': ARG_TYPE.STR,
        'description': 'API Key For Connecting to CoinBase API.',
        'required': True,
        'label': 'API Key',
        'secret': True
    },
    api_secret={
        'type': ARG_TYPE.PWD,
        'description': 'API Secret For Connecting to CoinBase API.',
        'required': True,
        'label': 'API Secret',
        'secret': True
    },
    api_passphrase={
        'type': ARG_TYPE.PWD,
        'description': 'API Passphrase.',
        'required': True,
        'label': 'API Passphrase',
        'secret': True
    },
)

connection_args_example = OrderedDict(
    api_key='public_key',
    api_secret='secret_key',
    api_passphrase='passphrase'
)



import datetime as dt
import ast
import pytz

import pandas as pd

from mindsdb_sql_parser import ast

from mindsdb.integrations.handlers.email_handler.email_ingestor import EmailIngestor
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.utilities.handlers.query_utilities.insert_query_utilities import INSERTQueryParser
from mindsdb.integrations.handlers.email_handler.settings import EmailSearchOptions
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class EmailsTable(APITable):
    """The Emails Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls email data from the connected account.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Emails matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'emails',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        for op, arg1, arg2 in where_conditions:
            if arg2 is None:
                logger.warning(f"Skipping condition: {arg1} {op} {arg2}."
                               "Please ignore if this is intentional, e.g. 'id > last' on first query of job run."
                               )
                continue

            if arg1 == 'datetime':
                date = self.parse_date(arg2)
                if op == '>':
                    search_params['since_date'] = date
                elif op == '<':
                    search_params['until_date'] = date
                else:
                    raise NotImplementedError("Only > and < operators are supported for created_at column.")
                continue

            elif arg1 == 'id':
                if op not in ['=', '>', '>=']:
                    raise NotImplementedError("Only =, > and >= operators are supported for id column.")
                if op in ['=', '>=']:
                    search_params['since_email_id'] = int(arg2)
                elif op == '>':
                    search_params['since_email_id'] = int(arg2) + 1

            elif arg1 in ['mailbox', 'subject', 'to_field', 'from_field']:
                if op != '=':
                    raise NotImplementedError("Only = operator is supported for mailbox, subject, to and from columns.")
                else:
                    if arg1 == 'from_field':
                        search_params['from_field'] = arg2
                    else:
                        search_params[arg1] = arg2

            else:
                raise NotImplementedError(f"Unsupported column: {arg1}.")

        self.handler.connect()

        if search_params:
            search_options = EmailSearchOptions(**search_params)
        else:
            search_options = EmailSearchOptions()

        email_ingestor = EmailIngestor(self.handler.connection, search_options)

        emails_df = email_ingestor.ingest()

        # ensure all queries from query are applied to the dataframe
        select_statement_executor = SELECTQueryExecutor(
            emails_df,
            selected_columns,
            [],
            order_by_conditions,
            result_limit
        )
        return select_statement_executor.execute_query()

    def insert(self, query: ast.Insert) -> None:
        """Sends emails through the connected account.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['to_field', 'subject', 'body'],
            mandatory_columns=['to_field', 'subject', 'body'],
            all_mandatory=True
        )
        email_data = insert_statement_parser.parse_query()

        for email in email_data:
            connection = self.handler.connect()
            to_addr = email['to_field']
            del email['to_field']
            connection.send_email(to_addr, **email)

    def get_columns(self):
        return ['id', 'body', 'subject', 'to_field', 'from_field', 'datetime']

    @staticmethod
    def parse_date(date_str) -> dt.datetime:

        if isinstance(date_str, dt.datetime):

            return date_str
        date_formats = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d']
        date = None
        for date_format in date_formats:
            try:
                date = dt.datetime.strptime(date_str, date_format)
            except ValueError:
                pass
        if date is None:
            raise ValueError(f"Can't parse date: {date_str}")
        date = date.astimezone(pytz.utc)

        return date



from dataclasses import dataclass


@dataclass(frozen=True)
class BaseConfig:
    time_limit: int = None
    num_cpus: int = 'auto'


@dataclass(frozen=True)
class ClassificationConfig(BaseConfig):
    pass


@dataclass(frozen=True)
class RegressionConfig(BaseConfig):
    pass



from typing import Optional, Dict
import pandas as pd
import requests

from monkeylearn import MonkeyLearn

from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.integrations.utilities.handler_utils import get_api_key


class monkeylearnHandler(BaseMLEngine):
    name = "monkeylearn"

    @staticmethod
    def create_validations(self,args=None,**kwargs):

        if "using" in args:
            args = args["using"]

        if "monkeylearn_api_key" not in args:
            raise Exception("monkeylearn_api_key not found")
        api_key = get_api_key('monkeylearn', args["using"], self.engine_storage, strict=False)
        if "model_id" in args:
            if "cl_" not in args["model_id"]:
                raise Exception("Classifier tasks are only supported currently")
        else:
            raise Exception("Enter the model_id of model you want use")
        model_id = args["model_id"]
        # Check whether the model_id given by user exists in the user account or monkeylearn pre-trained models
        url = 'https://api.monkeylearn.com/v3/classifiers/'
        response = requests.get(url, headers={'Authorization': 'Token {}'.format(api_key)})
        if response.status_code == 200:
            models = response.json()
            models_list = [model['id'] for model in models]
        else:
            raise Exception(f"Server response {response.status_code}")

        if model_id not in models_list:
            raise Exception(f"Model_id {args['model_id']} not found in MonkeyLearn pre-trained models")

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if "using" in args:
            args = args['using']

        self.model_storage.json_set('args', args)

    def predict(self, df, args=None):
        args = self.model_storage.json_get('args')
        input_column = args['input_column']
        if input_column not in df.columns:
            raise RuntimeError(f"input columns {input_column} not found ")
        input_list = df[input_column]
        if len(input_list) > 500:
            raise Exception("Classifier only supports 500 data elements in list")
        ml = get_api_key('monkeylearn', args["using"], self.engine_storage, strict=False)
        df_list = []
        for text in input_list:
            pred_dict = {}
            classifier_response = ml.classifiers.classify(args['model_id'], [text])
            for res_dict in classifier_response.body:
                if res_dict.get("error") is True:
                    raise Exception(res_dict["error_detail"])
                pred_dict['classification'] = res_dict['classifications']
                pred_dict['tag'] = res_dict['classifications'][0]['tag_name']
                df_list.append(pd.DataFrame([pred_dict]))
        pred_df = pd.concat(df_list)
        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        ml = get_api_key('monkeylearn', args["using"], self.engine_storage, strict=False)
        response = ml.classifiers.detail(args['model_id'])
        description = {}
        description['name'] = response.body['name']
        description['model_version'] = response.body['model_version']
        description['date_created'] = response.body['created']
        # pre-trained monkeylearn models guide about what industries they can be used
        description['industries'] = response.body['industries']
        des_df = pd.DataFrame([description])
        return des_df


from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    location={
        "type": ARG_TYPE.STR,
        "description": "If `:memory:` - use in-memory Qdrant instance. If a remote URL - connect to a remote Qdrant instance. Example: `http://localhost:6333`",
        "required": False,
    },
    url={
        "type": ARG_TYPE.STR,
        "description": "URL of Qdrant service. Either host or a string of type [scheme]<host><[port][prefix]. Ex: http://localhost:6333/service/v1",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "Host name of Qdrant service. The port and host are used to construct the connection URL.",
        "required": False,
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Port of the REST API interface. Default: 6333",
        "required": False,
    },
    grpc_port={
        "type": ARG_TYPE.INT,
        "description": "Port of the gRPC interface. Default: 6334",
        "required": False,
    },
    prefer_grpc={
        "type": ARG_TYPE.BOOL,
        "description": "If `true` - use gPRC interface whenever possible in custom methods. Default: false",
        "required": False,
    },
    https={
        "type": ARG_TYPE.BOOL,
        "description": "If `true` - use https protocol.",
        "required": False,
    },
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "API key for authentication in Qdrant Cloud.",
        "required": False,
        "secret": True
    },
    prefix={
        "type": ARG_TYPE.STR,
        "description": "If set, the value is added to the REST URL path. Example: `service/v1` will result in `http://localhost:6333/service/v1/{qdrant-endpoint}` for REST API",
        "required": False,
    },
    timeout={
        "type": ARG_TYPE.INT,
        "description": "Timeout for REST and gRPC API requests. Defaults to 5.0 seconds for REST and unlimited for gRPC",
        "required": False,
    },
    path={
        "type": ARG_TYPE.STR,
        "description": "Persistence path for a local Qdrant instance.",
        "required": False,
    },
    collection_config={
        "type": ARG_TYPE.DICT,
        "description": "Collection creation configuration. See https://qdrant.github.io/qdrant/redoc/index.html#tag/collections/operation/create_collection",
        "required": True,
    },
)

connection_args_example = {
    "location": ":memory:",
    "collection_config": {
        "size": 386,
        "distance": "Cosine"
    }
}



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.STR,
        "description": "The API key that can be found in your pinecone account",
        "required": True,
        "secret": True
    },
    environment={
        "type": ARG_TYPE.STR,
        "description": "The environment name corresponding to the `api_key`",
        "required": True,
    },
    dimension={
        "type": ARG_TYPE.INT,
        "description": "dimensions of the vectors to be stored in the index (default=8)",
        "required": False,
    },
    metric={
        "type": ARG_TYPE.STR,
        "description": "distance metric to be used for similarity search (default='cosine')",
        "required": False,
    },
    pods={
        "type": ARG_TYPE.INT,
        "description": "number of pods for the index to use, including replicas (default=1)",
        "required": False,
    },
    replicas={
        "type": ARG_TYPE.INT,
        "description": "the number of replicas. replicas duplicate your index. they provide higher availability and throughput (default=1)",
        "required": False,
    },
    pod_type={
        "type": ARG_TYPE.STR,
        "description": "the type of pod to use, refer to pinecone documentation (default='p1')",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    api_key="00000000-0000-0000-0000-000000000000",
    environment="gcp-starter",
    dimension=8,
    metric="cosine",
    pods=1,
    replicas=1,
    pod_type='p1',
)
